package encapsulation;

/**
 * Access class, represents an access from code to some entity in the program
 * like a method or a field
 * @author Janina Voigt
 *
 */
public class Access {

	private ExecutableBlock block;		//block where the access came from
	private AccessibleMember accessedMember;	//member that was accessed
	private String accessString;		//string of code used for access
	
	/**
	 * Constructor for Access objects
	 * @param theAccessedEntity the member that was accessed
	 * @param theBlock the block where the access came from
	 * @param theAccessString the string of code used for the access
	 */
	public Access(AccessibleMember theAccessedEntity, ExecutableBlock theBlock, String theAccessString){
		accessedMember = theAccessedEntity;
		block = theBlock;
		accessString = theAccessString;
	}
	
	/**
	 * Returns boolean indicating if the access was a method access
	 * @return true if the access was a method access, false otherwise
	 */
	public boolean isMethodAccess() {
		return accessedMember.isMethod();
	}
	
	/**
	 * Returns boolean indicating if the access was a field access
	 * @return true if the access was a field access, false otherwise
	 */
	public boolean isFieldAccess() {
		return accessedMember.isField();
	}
	
	/**
	 * Returns boolean indicating if the access came from within the object
	 * @return true if the access came from within the object, false otherwise
	 */
	public boolean isSameObject() {	
    	if(accessString.contains(".") && !accessString.startsWith("this") && !accessString.startsWith("super")){
			return false;
		}
    	return true;
    }
	
	/**
	 * Returns boolean indicating if the access came from within the class
	 * @return true if the access came from within the class, false otherwise
	 */
	public boolean isSameClass() {
		return accessedMember.getOwnerClass().equals(block.getOwnerClass());
	}
	
	/**
	 * Returns boolean indicating if the accessed member is public
	 * @return true if the accessed member is public, false otherwise
	 */
	public boolean isPublic() {
		return accessedMember.isPublic();
	}
	
	/**
	 * Returns boolean indicating if the accessed member is protected
	 * @return true if the accessed member is protected, false otherwise
	 */
	public boolean isProtected() {
		return accessedMember.isProtected();
	}
	
	/**
	 * Returns boolean indicating if the accessed member is package
	 * @return true if the accessed member is package, false otherwise
	 */
	public boolean isPackage() {
		return accessedMember.isPackage();
	}
	
	/**
	 * Returns boolean indicating if the accessed member is private
	 * @return true if the accessed member is private, false otherwise
	 */
	public boolean isPrivate() {
		return accessedMember.isPrivate();
	}
}
