package encapsulation;

import java.util.*;

import encapsulation.Access;

/**
 * AccessMetricCalculator class, represents a metrics calculator that calculates
 * metrics about member accesses
 * @author Janina Voigt
 *
 */
public class AccessMetricCalculator extends MetricCalculator {
	
	//sets containing accesses and members used for calculation
	private Set<Access> allAccesses;			
	private Set<AccessibleMember> allMembers;
	
	private Set<Access> otherObject;
	private Set<Access> otherClass;
	private Set<Access> publicAccess;
	private Set<Access> protectedAccess;
	private Set<Access> packageAccess;
	private Set<Access> privateAccess;
	
	private AccessStrategy strategy;	//strategy to extract relevant accesses and members
	
	/**
	 * Constructor for AccessMetricCalculator objects
	 * @param theEntity the entity to calculate the metric for
	 * @param theStrategy the strategy to use to extract relevant accesses
	 * and members
	 */
	public AccessMetricCalculator(Entity theEntity, AccessStrategy theStrategy) {		
		super(theEntity);
		strategy = theStrategy;
		
		allMembers = strategy.getMembers(theEntity);
		allAccesses = strategy.getAccesses(theEntity);
		
		otherObject = new HashSet<Access>();
		otherClass = new HashSet<Access>();
		publicAccess = new HashSet<Access>();
		protectedAccess = new HashSet<Access>();
		packageAccess = new HashSet<Access>();
		privateAccess = new HashSet<Access>();
	}
	
	/**
	 * Helper method to load the accesses into sets for easier calculation
	 * of metrics
	 */
	private void loadResults(){
		for(Access theAccess : allAccesses){
			if(!theAccess.isSameClass()){
				otherClass.add(theAccess);
			}
			if(!theAccess.isSameObject()){
				otherObject.add(theAccess);
			}
			if(theAccess.isPublic()){
				publicAccess.add(theAccess);
			}
			else if(theAccess.isProtected()){
				protectedAccess.add(theAccess);
			}
			else if(theAccess.isPackage()){
				packageAccess.add(theAccess);
			}
			else {
				privateAccess.add(theAccess);
			}
		}
	}

	/**
	 * Calculate the metric results and write them out using a ResultsWriter
	 * @param writer the ResultWriter to use to write out the metric results
	 */
	public void calculateResults(ResultsWriter writer){
		loadResults();
		
		List<Measurement> results = new ArrayList<Measurement>();
		
		results.add(new Measurement("Total accesses", String.valueOf(allAccesses.size()), entity));
		if(allAccesses.size() > 0){
			results.add(new Measurement("Total accesses per member", String.valueOf((float)allAccesses.size() / allMembers.size()), entity));
			results.add(new Measurement("Total accesses from outside the class", String.valueOf(otherClass.size()), entity));
			results.add(new Measurement("Total accesses from outside the object", String.valueOf(otherObject.size()), entity));
			
			int minimum = getMinimumAccesses();
			int maximum = getMaximumAccesses();
			results.add(new Measurement("Minimum number of accesses per member", String.valueOf(minimum), entity));
			results.add(new Measurement("Maximum number of accesses per member", String.valueOf(maximum), entity));
			
			results.add(new Measurement("Public", String.valueOf(publicAccess.size()), entity));
			results.add(new Measurement("Protected", String.valueOf(protectedAccess.size()), entity));
			results.add(new Measurement("Package", String.valueOf(packageAccess.size()), entity));
			results.add(new Measurement("Private", String.valueOf(privateAccess.size()), entity));
			
			Set<Access> tempAccesses = new HashSet<Access>(publicAccess);
			tempAccesses.retainAll(otherClass);
			results.add(new Measurement("Accesses to public members from outside the class", String.valueOf(tempAccesses.size()), entity));
			tempAccesses = new HashSet<Access>(publicAccess);
			tempAccesses.retainAll(otherObject);
			results.add(new Measurement("Accesses to public members from outside the object", String.valueOf(tempAccesses.size()), entity));
				
			tempAccesses = new HashSet<Access>(protectedAccess);
			tempAccesses.retainAll(otherClass);			
			results.add(new Measurement("Accesses to protected members from other classes", String.valueOf(tempAccesses.size()), entity));
			tempAccesses = new HashSet<Access>(protectedAccess);
			tempAccesses.retainAll(otherObject);
			results.add(new Measurement("Accesses to protected members from another object", String.valueOf(tempAccesses.size()), entity));
			
			tempAccesses = new HashSet<Access>(packageAccess);
			tempAccesses.retainAll(otherClass);
			results.add(new Measurement("Accesses to package members from outside the class", String.valueOf(tempAccesses.size()), entity));
			tempAccesses = new HashSet<Access>(packageAccess);
			tempAccesses.retainAll(otherObject);
			results.add(new Measurement("Accesses to package members from outside the object", String.valueOf(tempAccesses.size()), entity));
						
			tempAccesses = new HashSet<Access>(privateAccess);
			tempAccesses.retainAll(otherObject);
			results.add(new Measurement("Accesses to private members of another object of the same class", String.valueOf(tempAccesses.size()), entity));
		}
		writer.writeOutHeader(strategy.getHeader(entity));
		writer.writeOutResults(results);
	}
	
	/**
	 * Helper method to calculate the minimum accesses per entity
	 * @return the minimum accesses per entity
	 */
	private int getMinimumAccesses() {
		int minimum = -1;
		for(AccessibleMember member : allMembers){
			int count = member.getNumberOfAccesses();
			if(count < minimum || minimum == -1){
				minimum = count;
			}
		}
		return minimum;
	}
	
	/**
	 * Helper method to calculate the maximum accesses per entity
	 * @return the maximum accesses per entity
	 */
	private int getMaximumAccesses() {
		int maximum = -1;
		for(AccessibleMember member : allMembers){
			int count = member.getNumberOfAccesses();
			if(count > maximum){
				maximum = count;
			}
		}
		return maximum;
	}
}
