package encapsulation;

import java.util.*;

import jst.symtab.*;

/**
 * ClassOrInterface class, represents a class or an interface
 * @author Janina Voigt
 *
 */
public class ClassOrInterface implements Entity {
	
	private UserType userType;	//the JST UserType corresponding to this class or interface
	private Set<Field> fields;	//the fields of this class or interface
	private Set<Method> methods; 	//the methods of this class or interface
	private Set<ExecutableBlock> blocks;	//the blocks of this class or interface
	
	/**
	 * Constructor for ClassOrInterface objects
	 * @param theUserType the JST UserType corresponding to this class or interface
	 */
	public ClassOrInterface(UserType theUserType) {
		userType = theUserType;
		fields = new HashSet<Field>();
		methods = new HashSet<Method>();
		blocks = new HashSet<ExecutableBlock>();
	}
	
	/**
	 * Gets the name of the entity
	 * @return the name of the entity
	 */
	public String getName() {
		return userType.getSimpleName();
	}
	
	/**
	 * Adds a field to this class or interface
	 * @param field the field to add
	 */
	public void addField(Field field) {
		fields.add(field);
	}
	
	/**
	 * Adds a method to this class or interface
	 * @param method the method to add
	 */
	public void addMethod(Method method) {
		methods.add(method);
	}
	
	/**
	 * Adds a block to this class or interface
	 * @param block the block to add
	 */
	public void addBlock(ExecutableBlock block) {
		blocks.add(block);
	}

	/**
	 * Gets all members of this class or interface
	 * @return all members of this class or interface
	 */
	public Set<Member> getMembers(){
		Set<Member> allMembers = new HashSet<Member>(fields);
		allMembers.addAll(methods);
		allMembers.addAll(blocks);
		return allMembers;
	}
	
	/**
	 * Gets all accesses to members of this class or interface
	 * @return all accesses to members
	 */
	public Set<Access> getAccesses(){
		Set<Access> accesses = new HashSet<Access>();
		for(ExecutableBlock block : blocks){
			accesses.addAll(block.getAccesses());
		}
		return accesses;
	}
	
	/**
	 * Gets the AccessibleMember with the given JST decl
	 * @param decl the JST decl to search for
	 * @return the AccessibleMember with the same JST decl, or null if no 
	 * AccessibleMember was found
	 */
	public AccessibleMember getAccessibleMember(Decl decl){
		for(Method method : methods){
			if(method.isSameDecl(decl)){
				return method;
			}
		}
		for(Field field : fields){
			if(field.isSameDecl(decl)){
				return field;
			}
		}
		return null;
	}
}
