package encapsulation;

import java.util.*;

/**
 * FieldAccessStrategy class, defines a strategy for extracting accesses for metrics
 * calculations given an entity. This class extracts only field accesses from the 
 * given entity.
 * @author Janina Voigt
 *
 */
public class FieldAccessStrategy extends AccessStrategy {

	/**
	 * Gets the members that should be used in a metric calculation given a 
	 * certain entity
	 * @param entity the entity to get members from
	 * @return the members to use in the metrics calculation
	 */
	public Set<AccessibleMember> getMembers(Entity entity){
		Set<Member> allMembers = entity.getMembers();
		Set<AccessibleMember> members = new HashSet<AccessibleMember>();
		for(Member member : allMembers){
			if(member.isField()){
				members.add(member.asAccessibleMember());
			}
		}
		return members;
	}
	
	/**
	 * Gets the accesses that should be used in a metric calculation given a 
	 * certain entity
	 * @param entity the entity to get accesses from
	 * @return the accesses to use in the metrics calculation
	 */
	public Set<Access> getAccesses(Entity entity){
		Set<Access> allAccesses = entity.getAccesses();
		Set<Access> accesses = new HashSet<Access>();
		for(Access access : allAccesses){
			if(access.isFieldAccess()){
				accesses.add(access);
			}
		}
		return accesses;
	}
	
	/**
	 * Gets a header to describe this certain metric
	 * @param entity the entity used in the metric calculation
	 * @return a header to print
	 */
	public String getHeader(Entity entity){
		return "Field accesses for " + entity.getName();
	}
	
}
