package encapsulation;

import jst.symtab.*;

import java.util.*;

/**
 * Member class, represents a program entity that is within a class or interface
 * @author Janina Voigt
 *
 */
public abstract class Member implements Entity {
	
	protected ClassOrInterface owner;	//the class or interface that owns the member
	protected Set<Access> accesses;	//the accesses to this member
	
	/**
	 * Constructor for Member objects
	 * @param theOwner the owner class or interface
	 */
	public Member(ClassOrInterface theOwner) {
		owner = theOwner;
	}
	
	/**
	 * Gets the JST Decl corresponding to this member
	 * @return the JST Decl
	 */
	public abstract Decl getDecl();
	
	/**
	 * Returns the member as an AccessibleMember
	 * @return an AccessibleMember
	 */
	public AccessibleMember asAccessibleMember() {
		return null;
	}
	
	/**
	 * Gets the members within this member
	 * @return the members within this member
	 */
	public Set<Member> getMembers() {
		return new HashSet<Member>();
	}
	
	/**
	 * Gets the accesses to this member
	 * @return the accesses to this member
	 */
	public Set<Access> getAccesses() {
		return accesses;
	}

	/**
	 * Returns boolean indicating if the member is a method 
	 * @return true if the member is a method, false otherwise
	 */
	public boolean isMethod() {
		return false;
	}
	
	/**
	 * Returns boolean indicating if the field is a method 
	 * @return true if the member is a field, false otherwise
	 */
	public boolean isField() {
		return false;
	}
	
	/**
	 * Gets the number of accesses to this member
	 * @return the number of accesses to this member
	 */
	public int getNumberOfAccesses() {
		return accesses.size();
	}
	
	/**
	 * Adds an access to this member
	 * @param theAccess the new access
	 */
	public void addAccess(Access theAccess){
		accesses.add(theAccess);
	}
	
	/**
	 * Gets the name of the entity
	 * @return the name of the entity
	 */
	public String getName() {
		return getDecl().getSimpleName();
	}
	
	/**
	 * Gets the owner class for this member
	 * @return the owner class
	 */
	public ClassOrInterface getOwnerClass(){
		return owner;
	}
	
	/**
	 * Returns a boolean indicating if this member's JST decl is the same
	 * as a given JST decl
	 * @param decl the JST decl
	 * @return true if this member's JST decl is the same as the given decl,
	 * false otherwise
	 */
	public boolean isSameDecl(Decl decl) {
		return getDecl().equals(decl);
	}
}
