package encapsulation;

import java.util.*;

import encapsulation.AccessibleMember;

/**
 * MemberMetricCalculator class, represents a metrics calculator that calculates
 * metrics about members
 * @author Janina Voigt
 *
 */
public class MemberMetricCalculator extends MetricCalculator {

	//sets containing members used for calculation
	private Set<AccessibleMember> members;
	private Set<AccessibleMember> publicMembers;
	private Set<AccessibleMember> protectedMembers;
	private Set<AccessibleMember> packageMembers;
	private Set<AccessibleMember> privateMembers;
	
	private MemberStrategy strategy;	//strategy to extract relevant members
	
	/**
	 * Constructor for MemberMetricCalculator objects
	 * @param theEntity the entity to calculate the metric for
	 * @param theStrategy the strategy to use to extract relevant accesses
	 * and members
	 */
	public MemberMetricCalculator(Entity theEntity, MemberStrategy theStrategy){
		super(theEntity);
		strategy = theStrategy;
		
		members = strategy.getMembers(entity);
		
		publicMembers = new HashSet<AccessibleMember>();
		protectedMembers = new HashSet<AccessibleMember>();
		packageMembers = new HashSet<AccessibleMember>();
		privateMembers = new HashSet<AccessibleMember>();		
	}
	
	/**
	 * Helper method to load the accesses into sets for easier calculation
	 * of metrics
	 */
	private void loadResults() {
		for(AccessibleMember member : members){
			if(member.isPublic()){
				publicMembers.add(member);
			}
			else if(member.isProtected()){
				protectedMembers.add(member);
			}
			else if(member.isPackage()){
				packageMembers.add(member);
			}
			else {
				privateMembers.add(member);
			}
		}
	}
	
	/**
	 * Calculate the metric results and write them out using a ResultsWriter
	 * @param writer the ResultWriter to use to write out the metric results
	 */
	public void calculateResults(ResultsWriter writer) {
		loadResults();
		
		List<Measurement> results = new ArrayList<Measurement>();
		
		//print information about entities
		results.add(new Measurement("Total Count", String.valueOf(members.size()), entity));
		
		if(members.size() > 0) {
			results.add(new Measurement("Public", String.valueOf(publicMembers.size()), entity));
			results.add(new Measurement("Protected", String.valueOf(protectedMembers.size()), entity));
			results.add(new Measurement("Package", String.valueOf(packageMembers.size()), entity));
			results.add(new Measurement("Private", String.valueOf(privateMembers.size()), entity));
		}
		writer.writeOutHeader(strategy.getHeader(entity));
		writer.writeOutResults(results);
	}
}
