package encapsulation;

import java.io.BufferedWriter;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.SortedSet;
import java.util.TreeSet;

import jst.parsetree.DeclarationVisitor;
import jst.parsetree.ParseTree;
import jst.symtab.PackageDecl;
import jst.symtab.SourceFile;
import jst.symtab.visitor.ImportResolver;
import jst.symtab.visitor.MemberTypeResolver;
import jst.symtab.visitor.ReferenceResolver;
import jst.symtab.visitor.SuperInterfaceResolver;
import jst.symtab.visitor.SuperclassResolver;
import jst.symtab.visitor.XMLWriter;

/**
 * ParsetreeLoader class, represents an object that can read XMl parsetrees from
 * file and build a model of the program in memory
 */
public class ParsetreeLoader {
	
	public static final String USAGE =	"Usage: java encapsulation.Main [-c | -s] <source dir> [<source dir> ...] <new source dir>";
	
	/**
	 * Visit the model of the program using the given visitor
	 * @param visitor the visitor to use to visit the model
	 */
	public void visitModel(JSTModelVisitor visitor){
		PackageDecl defaultPackage = PackageDecl.getDefaultPackage();
		defaultPackage.accept(visitor);
	}

	/** 
	 *  Run the application.
     *  Load each parse tree, populate the symbol table, lookup references, then emit xml.
     */
    public void loadParsetrees(String arg) {
		System.err.println("Reading parse trees...");
		readSource(new File(arg));
        resolve();
        //save();
    }

    /** 
     *  Walk directory recursively, looking for xml files to read.
     *  Anything ending in ".xml" is assumed to contain a parse tree.
     *  Trees are loaded into the parseTrees array.
     */
    public void readSource(File file) {
        if (!file.exists()) {
            System.err.println("File not found: " + file);
            return;
        }

		// If this is a directory, walk each file/dir in that directory
		if (file.isDirectory()) {
			String files[] = file.list();
			for (int i = 0; i < files.length; i++) {
				File next = new File(file, files[i]);
				readSource(next);
			}
            return;
		}

		// Is this an xml file?
		if ((file.getName().length() <= 4) ||
			!file.getName().substring(file.getName().length() - 4).equals(".xml")) {
            return; // Not a ".xml" file.  Ignore it.
        }

		// Read the parse tree into memory.
        System.err.println("Reading " + file.getAbsolutePath());
		ParseTree tree = new ParseTree(file);
        addParseTree(tree);
    }

    /**
     * Add a parsetree to the model of the program
     * @param tree
     */
    public void addParseTree(ParseTree tree) {
        SourceFile sourceFile = new SourceFile(tree);
        // sourceFiles.add(sourceFile);

		// Find declarations in the parse tree.
        DeclarationVisitor visitor = new DeclarationVisitor(sourceFile);
        tree.getRoot().accept(visitor);
    }

    /**
     * Resolve the references in the model
     */
    public void resolve() {
		System.err.println("Resolving...");
    	SortedSet<String> messages = new TreeSet<String>();

		ImportResolver importResolver = new ImportResolver();
		PackageDecl.getDefaultPackage().accept(importResolver);

		// for (JSTMessage mess: importLinker.getMessages())
		for (Object mess: importResolver.getMessages())
			messages.add(mess.toString());

		SuperclassResolver superclassResolver = new SuperclassResolver();
		PackageDecl.getDefaultPackage().accept(superclassResolver);

		for (Object mess: superclassResolver.getMessages())
			messages.add(mess.toString());
		
		SuperInterfaceResolver superInterfaceResolver = new SuperInterfaceResolver();
		PackageDecl.getDefaultPackage().accept(superInterfaceResolver);

		for (Object mess: superInterfaceResolver.getMessages())
			messages.add(mess.toString());

		MemberTypeResolver typeResolver = new MemberTypeResolver();
		PackageDecl.getDefaultPackage().accept(typeResolver);

		for (Object mess: typeResolver.getMessages())
			messages.add(mess.toString());

		System.err.println("Resolving refs...");
		ReferenceResolver refResolver = new ReferenceResolver();
		PackageDecl.getDefaultPackage().accept(refResolver);

		for (Object mess: refResolver.getMessages())
			messages.add(mess.toString());

		if (messages.isEmpty())
			System.err.println("OK");
		else {
			System.err.println(messages.size() + " errors:");

			for (String message: messages)
				System.err.println("  " + message);
		}
    }

    /**
     * Save the model as XML
     */
	protected void save() {
		System.err.println("Writing XML file...");
		PrintWriter out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(System.out)));

		out.println("<?xml version='1.0' encoding='iso-8859-1'?>");
		out.println("<!DOCTYPE program SYSTEM 'jst.dtd'>");
		out.println("<program>");
        out.println("<symbols>");
        XMLWriter xml = new XMLWriter(out);
        PackageDecl.getDefaultPackage().accept(xml);
        out.println("</symbols>");
    	out.println("</program>");

		out.close();
    }
}
