
package jst.parsetree;
import jst.symtab.*;

import java.lang.reflect.Modifier;
import parser.parsetree.*;

public class DeclarationVisitor extends ParseTreeVisitor {

	protected Builder builder;

	protected boolean typeIsClass;
	protected boolean typeIsInterface;
	protected boolean typeIsUserType;

	protected boolean declaringField;
	protected boolean declaringParameter;
	protected boolean declaringLocal;

	protected boolean declaringAnonConstructorArgs;

	protected boolean blockIsStatic;

	// State variables for the visitor as it walks down the tree.
	protected boolean declaringSuperInterfaces;

	protected int extraDims;

	protected SyntaxTreeBranch savedModifiers;
	protected SyntaxTreeBranch savedType;
	protected SyntaxTreeBranch savedVariableInitializer;
	
    public DeclarationVisitor(SourceFile sourceFile) {
        builder = new Builder(sourceFile);
    }

    public void visitBranch(SyntaxTreeBranch branch) {

		if (branch.isAmbiguous()) {
			visitAmbiguousBranch(branch);
			return;
		}

		assert getType(branch) <= Symbol.NT_GOAL : getType(branch);

        switch (getType(branch)) {

            case Symbol.NT_TYPE_DECL_SPECIFIER					: visitTypeDeclSpecifier(branch); return;
            case Symbol.NT_TYPE_VARIABLE						: visitTypeVariable(branch); return;
            case Symbol.NT_ARRAY_TYPE							: visitArrayType(branch); return;
            case Symbol.NT_TYPE_PARAMETER						: visitTypeParameter(branch); return;
            case Symbol.NT_TYPE_BOUND							: visitTypeBound(branch); return;
            case Symbol.NT_ADDITIONAL_BOUND						: visitAdditionalBound(branch); return;
            case Symbol.NT_ACTUAL_TYPE_ARGUMENT					: visitActualTypeArgument(branch); return;
            case Symbol.NT_WILDCARD								: visitWildCard(branch); return;
            case Symbol.NT_WILDCARD_BOUNDS						: visitWildCardBounds(branch); return;
            case Symbol.NT_PACKAGE_NAME							: visitPackageName(branch); return;
            case Symbol.NT_TYPE_NAME							: visitTypeName(branch); return;
            case Symbol.NT_EXPRESSION_NAME						: visitExpressionName(branch); return;
            case Symbol.NT_METHOD_NAME							: visitMethodName(branch); return;
            case Symbol.NT_PACKAGE_OR_TYPE_NAME					: visitPackageOrTypeName(branch); return;
            case Symbol.NT_AMBIGUOUS_NAME						: visitAmbiguousName(branch); return;
            case Symbol.NT_CLASS_NAME							: visitClassName(branch); return;
            case Symbol.NT_COMPILATION_UNIT						: visitCompilationUnit(branch); return;
            case Symbol.NT_PACKAGE_DECLARATION					: visitPackageDeclaration(branch); return;
            case Symbol.NT_SINGLE_TYPE_IMPORT_DECLARATION		: visitSingleTypeImportDeclaration(branch); return;
            case Symbol.NT_TYPE_IMPORT_ON_DEMAND_DECLARATION	: visitTypeImportOnDemandDeclaration(branch); return;
            case Symbol.NT_SINGLE_STATIC_IMPORT_DECLARATION		: visitSingleStaticImportDeclaration(branch); return;
            case Symbol.NT_STATIC_IMPORT_ON_DEMAND_DECLARATION	: visitStaticImportOnDemandDeclaration(branch); return;
            case Symbol.NT_NORMAL_CLASS_DECLARATION				: visitNormalClassDeclaration(branch); return;
            case Symbol.NT_SUPER								: visitSuper(branch); return;
            case Symbol.NT_INTERFACE_TYPE_LIST					: visitInterfaceTypeList(branch); return;
            case Symbol.NT_FIELD_DECLARATION 					: visitFieldDeclaration(branch); return;
            case Symbol.NT_VARIABLE_DECLARATOR					: visitVariableDeclarator(branch); return;
            case Symbol.NT_VARIABLE_DECLARATOR_ID				: visitVariableDeclaratorId(branch); return;
            case Symbol.NT_VARIABLE_INITIALIZER					: visitVariableInitializer(branch); return;
            case Symbol.NT_METHOD_DECLARATION					: visitMethodDeclaration(branch); return;
            case Symbol.NT_METHOD_HEADER						: visitMethodHeader(branch); return;
            case Symbol.NT_METHOD_DECLARATOR					: visitMethodDeclarator(branch); return;
            case Symbol.NT_FORMAL_PARAMETER						: visitFormalParameter(branch); return;
            case Symbol.NT_LAST_FORMAL_PARAMETER				: visitLastFormalParameter(branch); return;
            case Symbol.NT_EXCEPTION_TYPE						: visitExceptionType(branch); return;
            case Symbol.NT_STATIC_INITIALIZER					: visitStaticInitializer(branch); return;
            case Symbol.NT_CONSTRUCTOR_DECLARATION				: visitConstructorDeclaration(branch); return;
            case Symbol.NT_CONSTRUCTOR_DECLARATOR				: visitConstructorDeclarator(branch); return;
            case Symbol.NT_CONSTRUCTOR_BODY						: visitConstructorBody(branch); return;
            case Symbol.NT_EXPLICIT_CONSTRUCTOR_INVOCATION		: visitExplicitConstructorInvocation(branch); return;
            case Symbol.NT_REFERENCE_TYPE_LIST					: visitReferenceTypeList(branch); return;
            case Symbol.NT_ENUM_DECLARATION						: visitEnumDeclaration(branch); return;
            case Symbol.NT_ENUM_CONSTANT						: visitEnumConstant(branch); return;
            case Symbol.NT_NORMAL_INTERFACE_DECLARATION			: visitNormalInterfaceDeclaration(branch); return;
            case Symbol.NT_EXTENDS_INTERFACES					: visitExtendsInterfaces(branch); return;
            // case Symbol.NT_INTERFACE_BODY						: visitInterfaceBody(branch); return;
            case Symbol.NT_CONSTANT_DECLARATION					: visitConstantDeclaration(branch); return;
            case Symbol.NT_ABSTRACT_METHOD_DECLARATION			: visitAbstractMethodDeclaration(branch); return;
			case Symbol.NT_BLOCK								: visitBlock(branch); return;
			// case Symbol.NT_BLOCK_STATEMENT						: visitBlockStatement(branch); return;
            case Symbol.NT_LOCAL_VARIABLE_DECLARATION			: visitLocalVariableDeclaration(branch); return;
            case Symbol.NT_STATEMENT_EXPRESSION					: visitStatementExpression(branch); return;
            case Symbol.NT_IF_THEN_STATEMENT					: visitIfThenStatement(branch); return;
            case Symbol.NT_IF_THEN_ELSE_STATEMENT				: visitIfThenElseStatement(branch); return;
            case Symbol.NT_IF_THEN_ELSE_STATEMENT_NO_SHORT_IF	: visitIfThenElseStatementNoShortIf(branch); return;
            case Symbol.NT_ASSERT_STATEMENT						: visitAssertStatement(branch); return;
            case Symbol.NT_SWITCH_STATEMENT						: visitSwitchStatement(branch); return;
            case Symbol.NT_SWITCH_LABEL							: visitSwitchLabel(branch); return;
            case Symbol.NT_WHILE_STATEMENT						: visitWhileStatement(branch); return;
            case Symbol.NT_WHILE_STATEMENT_NO_SHORT_IF			: visitWhileStatementNoShortIf(branch); return;
            case Symbol.NT_DO_STATEMENT							: visitDoStatement(branch); return;
            case Symbol.NT_BASIC_FOR_STATEMENT					: visitBasicForStatement(branch); return;
            case Symbol.NT_FOR_STATEMENT_NO_SHORT_IF			: visitForStatementNoShortIf(branch); return;
            case Symbol.NT_ENHANCED_FOR_STATEMENT				: visitEnhancedForStatement(branch); return;
            case Symbol.NT_RETURN_STATEMENT						: visitReturnStatement(branch); return;
            case Symbol.NT_THROW_STATEMENT						: visitThrowStatement(branch); return;
            case Symbol.NT_SYNCHRONIZED_STATEMENT				: visitSynchronizedStatement(branch); return;
			case Symbol.NT_CATCH_CLAUSE							: visitCatchClause(branch); return;
			case Symbol.NT_FINALLY								: visitFinally(branch); return;
			case Symbol.NT_PRIMARY_NO_NEW_ARRAY					: visitPrimaryNoNewArray(branch); return;
			case Symbol.NT_CLASS_INSTANCE_CREATION_EXPRESSION	: visitClassInstanceCreationExpression(branch); return;
            case Symbol.NT_ARGUMENT_LIST						: visitArgumentList(branch); return;
			case Symbol.NT_ARRAY_CREATION_EXPRESSION			: visitArrayCreationExpression(branch); return;
			case Symbol.NT_DIM_EXPR								: visitDimExpr(branch); return;
			case Symbol.NT_DIMS									: visitDims(branch); return;
			case Symbol.NT_FIELD_ACCESS							: visitFieldAccess(branch); return;
            case Symbol.NT_METHOD_INVOCATION					: visitMethodInvocation(branch); return;
			case Symbol.NT_ARRAY_ACCESS							: visitArrayAccess(branch); return;
			case Symbol.NT_UNARY_EXPRESSION_NOT_PLUS_MINUS		: visitUnaryExpressionNotPlusMinus(branch); return;
			case Symbol.NT_CAST_EXPRESSION						: visitCastExpression(branch); return;
			case Symbol.NT_MULTIPLICATIVE_EXPRESSION			: visitMultiplicativeExpression(branch); return;
			case Symbol.NT_ADDITIVE_EXPRESSION					: visitAdditiveExpression(branch); return;
			case Symbol.NT_SHIFT_EXPRESSION						: visitShiftExpression(branch); return;
			case Symbol.NT_RELATIONAL_EXPRESSION				: visitRelationalExpression(branch); return;
			case Symbol.NT_EQUALITY_EXPRESSION					: visitEqualityExpression(branch); return;
			case Symbol.NT_AND_EXPRESSION						: visitAndExpression(branch); return;
			case Symbol.NT_EXCLUSIVE_OR_EXPRESSION				: visitExclusiveOrExpression(branch); return;
			case Symbol.NT_INCLUSIVE_OR_EXPRESSION				: visitInclusiveOrExpression(branch); return;
			case Symbol.NT_CONDITIONAL_AND_EXPRESSION			: visitConditionalAndExpression(branch); return;
			case Symbol.NT_CONDITIONAL_OR_EXPRESSION			: visitConditionalOrExpression(branch); return;
			case Symbol.NT_CONDITIONAL_EXPRESSION				: visitConditionalExpression(branch); return;
			case Symbol.NT_ASSIGNMENT							: visitAssignment(branch); return;
			case Symbol.NT_EXPRESSION 							: visitExpression(branch); return;

			// Walk through nodes by default.
            default												: super.visitBranch(branch); return;
        }
    }

	public void visitAmbiguousBranch(SyntaxTreeBranch branch) {

		switch (getType(branch)) {
			case Symbol.NT_REFERENCE_TYPE			: visitAmbiguousReferenceType(branch); return;
			case Symbol.NT_TYPE_DECL_SPECIFIER		: visitAmbiguousTypeDeclSpecifier(branch); return;
			case Symbol.NT_CLASS_OR_INTERFACE_TYPE	: visitAmbiguousClassOrInterfaceType(branch); return;
			case Symbol.NT_TYPE_BOUND				: visitAmbiguousTypeBound(branch); return;
            case Symbol.NT_EXCEPTION_TYPE			: visitAmbiguousExceptionType(branch); return;
            default									: branch.getKids()[0].accept(this); return;
		}
	}

	protected void visitAmbiguousReferenceType(SyntaxTreeBranch ambigReferenceType) {

		// Ambiguity is between ClassOrInterfaceType and TypeVariable.
		// Just ignore TypeVariable, and create a normal TypeLookup. It can find type variables.
		SyntaxTreeBranch alt1 = (SyntaxTreeBranch) ambigReferenceType.getKids()[0];
		SyntaxTreeBranch alt2 = (SyntaxTreeBranch) ambigReferenceType.getKids()[1];
		
		SyntaxTreeBranch classOrInterfaceType = getBranch(alt1, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 0, 0);
		if (classOrInterfaceType == null)
			classOrInterfaceType = getBranch(alt2, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 0, 0);

		classOrInterfaceType.accept(this);
	}

	protected void visitAmbiguousClassOrInterfaceType(SyntaxTreeBranch ambigClassOrInterfaceType) {
		// Both alternatives produce the same lookup, so just do one.
		SyntaxTreeBranch alt1 = (SyntaxTreeBranch) ambigClassOrInterfaceType.getKids()[0];
		alt1.accept(this);
	}

	protected void visitAmbiguousTypeDeclSpecifier(SyntaxTreeBranch ambigTypeDeclSpecifier) {

		// Neither alternative has TypeArguments.
		// The <TypeName> production is more general.
		// It semantically allows TypeVariable lookup, whereas
		// the alternative allows only UserTypes.
		SyntaxTreeBranch alt1 = (SyntaxTreeBranch) ambigTypeDeclSpecifier.getKids()[0];
		SyntaxTreeBranch alt2 = (SyntaxTreeBranch) ambigTypeDeclSpecifier.getKids()[1];
		
		SyntaxTreeBranch typeName = getBranch(alt1, Symbol.NT_TYPE_NAME, 0, 0);
		if (typeName == null)
			typeName = getBranch(alt2, Symbol.NT_TYPE_NAME, 0, 0);

		typeName.accept(this);
	}

	protected void visitTypeDeclSpecifier(SyntaxTreeBranch typeDeclSpecifier) {
		SyntaxTreeBranch typeName = getBranch(typeDeclSpecifier, Symbol.NT_TYPE_NAME, 0, 0);
		if (typeName != null) {
			// typeIsUserType = true;
			typeName.accept(this);
			return;
		}

		SyntaxTreeBranch classOrInterfaceType = getBranch(typeDeclSpecifier, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 0, 0);

		boolean classType = typeIsClass;
		boolean interfaceType = typeIsInterface;
		typeIsClass = false;
		typeIsInterface = false;
		typeIsUserType = false;

		classOrInterfaceType.accept(this);

		SyntaxTreeLeaf id = getLeaf(typeDeclSpecifier, Symbol.T_IDENTIFIER, 2, 2);

		if (classType)
			builder.lookupClassType(id);
		else if (interfaceType)
			builder.lookupInterfaceType(id);
		else
			builder.lookupUserType(id);
	}

	protected void visitTypeVariable(SyntaxTreeBranch typeVariable) {
		assert false;

		// Type Variable lookups are always ambiguous, & we always choose the alternative.
	}

	protected void visitArrayType(SyntaxTreeBranch arrayType) {
		SyntaxTreeBranch type = getBranch(arrayType, Symbol.NT_TYPE, 0, 0);
		type.accept(this);
		builder.lookupArray();
	}

	protected void visitTypeParameter(SyntaxTreeBranch typeParameter) {
		SyntaxTreeBranch typeVariable = getBranch(typeParameter, Symbol.NT_TYPE_VARIABLE, 0, 0);
		builder.startTypeParameter(typeVariable.getValue(), typeParameter);

		SyntaxTreeBranch typeBound = getBranch(typeParameter, Symbol.NT_TYPE_BOUND, 1, 1);
		if (typeBound != null)
			typeBound.accept(this);

		builder.endTypeParameter();
	}

	protected void visitAmbiguousTypeBound(SyntaxTreeBranch ambigTypeBound) {

		// Ambiguity is between ClassOrInterfaceType and TypeVariable.
		// Just ignore TypeVariable, and create a normal TypeLookup. It can find type variables.
		SyntaxTreeBranch alt1 = (SyntaxTreeBranch) ambigTypeBound.getKids()[0];
		SyntaxTreeBranch alt2 = (SyntaxTreeBranch) ambigTypeBound.getKids()[1];
		
		SyntaxTreeBranch classOrInterfaceType = getBranch(alt1, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 1, 1);
		if (classOrInterfaceType != null)
			alt1.accept(this);
		else
			alt2.accept(this);
	}

	protected void visitTypeBound(SyntaxTreeBranch typeBound) {
		SyntaxTreeBranch type = getBranch(typeBound, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 1, 1);
		if (type == null)
			type =  getBranch(typeBound, Symbol.NT_TYPE_VARIABLE, 1, 1);

		type.accept(this);
		builder.setTypeBound();

		SyntaxTreeBranch additionalBoundList = getBranch(typeBound, Symbol.NT_ADDITIONAL_BOUND_LIST, 2, 2);
		if (additionalBoundList != null)
			additionalBoundList.accept(this);
	}

	protected void visitAdditionalBound(SyntaxTreeBranch additionalBound) {
		typeIsInterface = true;
		super.visitBranch(additionalBound);
		builder.setTypeBound();
	}

	protected void visitActualTypeArgument(SyntaxTreeBranch actualTypeArgument) {
		builder.startTypeArgument();
		super.visitBranch(actualTypeArgument);
		builder.endTypeArgument();
	}

	protected void visitWildCard(SyntaxTreeBranch wildCard) {
		builder.lookupWildcard(getLeaf(wildCard, Symbol.T_QUESTION, 0, 0));
		super.visitBranch(wildCard);
	}

	protected void visitWildCardBounds(SyntaxTreeBranch wildCardBounds) {
		builder.startBound();
		super.visitBranch(wildCardBounds);

		boolean upper = getLeaf(wildCardBounds, Symbol.T_EXTENDS, 0, 0) != null;
		builder.endBound(upper);
	}

	protected void visitPackageName(SyntaxTreeBranch packageName) {
		SyntaxTreeBranch leftPackageName = getBranch(packageName, Symbol.NT_PACKAGE_NAME, 0, 0);
		if (leftPackageName != null)
			leftPackageName.accept(this);

		SyntaxTreeLeaf id = getLeaf(packageName, Symbol.T_IDENTIFIER, 0, 2);
		builder.addPackageName(id.getValue(), packageName);
	}

	protected void visitTypeName(SyntaxTreeBranch typeName) {
		SyntaxTreeBranch packageOrTypeName = getBranch(typeName, Symbol.NT_PACKAGE_OR_TYPE_NAME, 0, 0);
		if (packageOrTypeName != null)
			packageOrTypeName.accept(this);

		SyntaxTreeLeaf id = getLeaf(typeName, Symbol.T_IDENTIFIER, 0, 2);

		if (typeIsUserType)
			builder.lookupUserType(id);
		else if (typeIsClass)
			builder.lookupClassType(id);
		else if (typeIsInterface)
			builder.lookupInterfaceType(id);
		else
			builder.lookupType(id); // Allows TypeVariables too.

		typeIsUserType = false;
		typeIsClass = false;
		typeIsInterface = false;
	}

	protected void visitExpressionName(SyntaxTreeBranch expressionName) {
		SyntaxTreeBranch ambiguousName = getBranch(expressionName, Symbol.NT_AMBIGUOUS_NAME, 0, 0);
		if (ambiguousName != null)
			ambiguousName.accept(this);

		SyntaxTreeLeaf id = getLeaf(expressionName, Symbol.T_IDENTIFIER, 0, 2);

		builder.lookupVariable(id);
	}

	protected void visitMethodName(SyntaxTreeBranch methodName) {
		SyntaxTreeBranch ambiguousName = getBranch(methodName, Symbol.NT_AMBIGUOUS_NAME, 0, 0);
		if (ambiguousName != null)
			ambiguousName.accept(this);

		SyntaxTreeLeaf id = getLeaf(methodName, Symbol.T_IDENTIFIER, 0, 2);

		builder.lookupMethod(id);
	}

	protected void visitPackageOrTypeName(SyntaxTreeBranch packageOrTypeName) {
		SyntaxTreeBranch leftPackageOrTypeName = getBranch(packageOrTypeName, Symbol.NT_PACKAGE_OR_TYPE_NAME, 0, 0);
		if (leftPackageOrTypeName != null)
			leftPackageOrTypeName.accept(this);

		SyntaxTreeLeaf id = getLeaf(packageOrTypeName, Symbol.T_IDENTIFIER, 0, 2);

		builder.lookupPackageOrType(id);
	}

	protected void visitAmbiguousName(SyntaxTreeBranch ambiguousName) {
		SyntaxTreeBranch leftAmbiguousName = getBranch(ambiguousName, Symbol.NT_AMBIGUOUS_NAME, 0, 0);
		if (leftAmbiguousName != null)
			leftAmbiguousName.accept(this);

		SyntaxTreeLeaf id = getLeaf(ambiguousName, Symbol.T_IDENTIFIER, 0, 2);

		builder.lookupAmbiguous(id);
	}

	protected void visitClassName(SyntaxTreeBranch className) {
		SyntaxTreeBranch packageOrTypeName = getBranch(className, Symbol.NT_PACKAGE_OR_TYPE_NAME, 0, 0);
		if (packageOrTypeName != null)
			packageOrTypeName.accept(this);

		SyntaxTreeLeaf id = getLeaf(className, Symbol.T_IDENTIFIER, 0, 2);
		builder.lookupClassType(id);
	}

	protected void visitCompilationUnit(SyntaxTreeBranch compilationUnit) {
		SyntaxTreeBranch packageDeclaration = getBranch(compilationUnit, Symbol.NT_PACKAGE_DECLARATION, 0, 0);
		if (packageDeclaration == null)
			builder.setPackage(); // Default pkg.

		super.visitBranch(compilationUnit);

	}

	protected void visitPackageDeclaration(SyntaxTreeBranch packageDeclaration) {
		super.visitBranch(packageDeclaration);

		builder.setPackage();
	}

	protected void visitSingleTypeImportDeclaration(SyntaxTreeBranch singleTypeImportDeclaration) {

		SyntaxTreeBranch typeName = getBranch(singleTypeImportDeclaration, Symbol.NT_TYPE_NAME, 1, 1);

		typeIsUserType = true;
		typeName.accept(this);

		builder.importType(false);
    }

	protected void visitTypeImportOnDemandDeclaration(SyntaxTreeBranch typeImportOnDemandDeclaration) {
		SyntaxTreeBranch packageOrTypeName = getBranch(typeImportOnDemandDeclaration, Symbol.NT_PACKAGE_OR_TYPE_NAME, 1, 1);
		packageOrTypeName.accept(this);

		builder.importOnDemand(false);
	}

	protected void visitSingleStaticImportDeclaration(SyntaxTreeBranch singleStaticImportDeclaration) {
		SyntaxTreeBranch typeName = getBranch(singleStaticImportDeclaration, Symbol.NT_TYPE_NAME, 2, 2);

		typeIsUserType = true;
		typeName.accept(this);

		builder.importType(true);
	}

	protected void visitStaticImportOnDemandDeclaration(SyntaxTreeBranch staticImportOnDemandDeclaration) {
		SyntaxTreeBranch typeName = getBranch(staticImportOnDemandDeclaration, Symbol.NT_TYPE_NAME, 2, 2);

		typeIsUserType = true;
		typeName.accept(this);

		builder.importOnDemand(true);
	}

	/**
     * Declare a class.
     */
	protected void visitNormalClassDeclaration(SyntaxTreeBranch normalClassDeclaration) {
		SyntaxTreeLeaf ident = getLeaf(normalClassDeclaration, Symbol.T_IDENTIFIER, 1, 2);
        builder.startClass(ident.getValue(), normalClassDeclaration);

		SyntaxTreeBranch zuper = getBranch(normalClassDeclaration, Symbol.NT_SUPER, 2, 4);
		if (zuper == null)
			builder.noSuper();

		super.visitBranch(normalClassDeclaration);
		builder.endClass();
	}

	protected void visitSuper(SyntaxTreeBranch zuper) {
		typeIsClass = true;
		super.visitBranch(zuper);
	
		builder.zuper();
	}

	protected void visitInterfaceTypeList(SyntaxTreeBranch interfaceTypeList) {

        SyntaxTreeBranch leftInterfaceTypeList = getBranch(interfaceTypeList, Symbol.NT_INTERFACE_TYPE_LIST, 0, 0);
		if (leftInterfaceTypeList != null)
			leftInterfaceTypeList.accept(this);

        SyntaxTreeBranch interfaceType = getBranch(interfaceTypeList, Symbol.NT_INTERFACE_TYPE, 0, 2);

		typeIsInterface = true;
		super.visitBranch(interfaceType);

		builder.superInterface();
	}

	// One field declaration might actually declare many fields.
	protected void visitFieldDeclaration(SyntaxTreeBranch fieldDeclaration) {

		// Don't visit type or modifiers until lower down, when we have actually declared a field.
		savedModifiers = getBranch(fieldDeclaration, Symbol.NT_FIELD_MODIFIERS,  0, 0);
		savedType = getBranch(fieldDeclaration, Symbol.NT_TYPE,  0, 1);

		SyntaxTreeBranch variableDeclarators = getBranch(fieldDeclaration, Symbol.NT_VARIABLE_DECLARATORS, 1, 2);

		declaringField = true;
		variableDeclarators.accept(this);
		declaringField = false;
		savedModifiers = null;
    }	

	protected void visitVariableDeclarator(SyntaxTreeBranch variableDeclarator) {
		savedVariableInitializer = getBranch(variableDeclarator, Symbol.NT_VARIABLE_INITIALIZER, 2, 2);

		SyntaxTreeBranch variableDeclaratorId = getBranch(variableDeclarator, Symbol.NT_VARIABLE_DECLARATOR_ID, 0, 0);
		variableDeclaratorId.accept(this);
	}

	protected void visitVariableDeclaratorId(SyntaxTreeBranch variableDeclaratorId) {
		SyntaxTreeBranch leftVariableDeclaratorId = getBranch(variableDeclaratorId, Symbol.NT_VARIABLE_DECLARATOR_ID, 0, 0);

		if (leftVariableDeclaratorId != null) {
			extraDims++;
			leftVariableDeclaratorId.accept(this);
			return;
		}

		SyntaxTreeLeaf id = getLeaf(variableDeclaratorId, Symbol.T_IDENTIFIER, 0, 0);
		
		if (declaringField)
			builder.startField(id.getValue(), variableDeclaratorId);
		else if (declaringParameter)
			builder.startParameter(id.getValue(), variableDeclaratorId);
		else if (declaringLocal)
			builder.startLocalVariable(id.getValue(), variableDeclaratorId);
		else
			assert false : "Invalid owner of VariableDeclaratorId: " + variableDeclaratorId;

		if (savedModifiers != null)
			savedModifiers.accept(this);

		savedType.accept(this);

		while (extraDims > 0) {
			builder.lookupArray();
			extraDims--;
		}

		builder.setType();

		// Initializers fall outside scope of whatever they are initializing, unless it is a field.
		if (declaringParameter)
			builder.endParameter();
		else if (declaringLocal)
			builder.endLocalVariable();

		if (savedVariableInitializer != null) {
			SyntaxTreeBranch initializer = savedVariableInitializer;
			savedVariableInitializer = null; // Might hit other decls inside the initializer.

			initializer.accept(this);

		}

		if (declaringField){
			builder.endField();
		}
	}

	protected void visitVariableInitializer(SyntaxTreeBranch variableInitializer) {
		// Oh, this is ugly.  We might encounter all sorts of stuff in the initializer, including whole classes,
		// so we need to preserve flags that might get overwritten.
		boolean field = declaringField;
		declaringField = false;

		boolean local = declaringLocal;
		declaringLocal = false;

		assert field || local;

		SyntaxTreeBranch type = savedType;
		savedType = null;

		SyntaxTreeBranch modifiers = savedModifiers;
		savedModifiers = null;

		if (field)
			builder.startFieldInitializer(variableInitializer);

		super.visitBranch(variableInitializer);
		builder.addReference();

		if (field)
			builder.endFieldInitializer();

		// Restore the pre-initializer state.
		declaringField = field;
		declaringLocal = local;
		savedType = type;
		savedModifiers = modifiers;
	}

	protected void visitMethodDeclaration(SyntaxTreeBranch methodDeclaration) {
		super.visitBranch(methodDeclaration);
		builder.endMethod(); // Delayed until after method body.
	}
	
	protected void visitMethodHeader(SyntaxTreeBranch methodHeader) {
		
		SyntaxTreeBranch methodDeclarator = getBranch(methodHeader, Symbol.NT_METHOD_DECLARATOR, 1, 3);
		methodDeclarator.accept(this);

		SyntaxTreeBranch methodModifiers = getBranch(methodHeader, Symbol.NT_METHOD_MODIFIERS, 0, 0);
		if (methodModifiers != null)
			methodModifiers.accept(this);

		SyntaxTreeBranch typeParameters = getBranch(methodHeader, Symbol.NT_TYPE_PARAMETERS, 0, 1);
		if (typeParameters != null)
			typeParameters.accept(this);

		SyntaxTreeBranch resultType = getBranch(methodHeader, Symbol.NT_RESULT_TYPE, 0, 2);
		resultType.accept(this);

		while (extraDims > 0) {
			builder.lookupArray();
			extraDims--;
		}

		builder.setType();

		SyntaxTreeBranch throze = getBranch(methodHeader, Symbol.NT_THROWS, 2, 4);
		if (throze != null)
			throze.accept(this);
	}

	protected void visitMethodDeclarator(SyntaxTreeBranch methodDeclarator) {
		SyntaxTreeBranch leftMethodDeclarator = getBranch(methodDeclarator, Symbol.NT_METHOD_DECLARATOR, 0, 0);

		if (leftMethodDeclarator != null) {
			extraDims++;
			leftMethodDeclarator.accept(this);
			return;
		}

		SyntaxTreeLeaf id = getLeaf(methodDeclarator, Symbol.T_IDENTIFIER, 0, 0);
		
		builder.startMethod(id.getValue(), methodDeclarator);

		SyntaxTreeBranch formalParameterList = getBranch(methodDeclarator, Symbol.NT_FORMAL_PARAMETER_LIST, 2, 2);
        if (formalParameterList != null)
            formalParameterList.accept(this);

		// Delay endMethod until after body.
	}

	protected void visitFormalParameter(SyntaxTreeBranch formalParameter) {

		// Don't visit type or modifiers until lower down, when we have actually declared a parameter.
		savedModifiers = getBranch(formalParameter, Symbol.NT_VARIABLE_MODIFIERS,  0, 0);
		savedType = getBranch(formalParameter, Symbol.NT_TYPE,  0, 1);

		SyntaxTreeBranch variableDeclaratorId = getBranch(formalParameter, Symbol.NT_VARIABLE_DECLARATOR_ID, 1, 2);

		declaringParameter = true;
		variableDeclaratorId.accept(this);
		declaringParameter = false;
    }

	protected void visitLastFormalParameter(SyntaxTreeBranch lastFormalParameter) {
		SyntaxTreeBranch formalParameter = getBranch(lastFormalParameter, Symbol.NT_FORMAL_PARAMETER, 0, 0);
		if (formalParameter != null) {
			formalParameter.accept(this);
			return;
		}

		// Don't visit type or modifiers until lower down, when we have actually declared a parameter.
		savedModifiers = getBranch(lastFormalParameter, Symbol.NT_VARIABLE_MODIFIERS,  0, 0);
		savedType = getBranch(lastFormalParameter, Symbol.NT_TYPE,  0, 1);

		SyntaxTreeBranch variableDeclaratorId = getBranch(lastFormalParameter, Symbol.NT_VARIABLE_DECLARATOR_ID, 2, 3);

		extraDims++; // Turn ... into an array.

		declaringParameter = true;
		variableDeclaratorId.accept(this);
		declaringParameter = false;
	}	

	protected void visitAmbiguousExceptionType(SyntaxTreeBranch ambigExceptionType) {

		// Ambiguity is between ClassOrInterfaceType and TypeVariable.
		// Just ignore TypeVariable, and create a normal TypeLookup. It can find type variables.
		SyntaxTreeBranch alt1 = (SyntaxTreeBranch) ambigExceptionType.getKids()[0];
		SyntaxTreeBranch alt2 = (SyntaxTreeBranch) ambigExceptionType.getKids()[1];
		
		SyntaxTreeBranch classType = getBranch(alt1, Symbol.NT_CLASS_TYPE, 0, 0);
		if (classType != null)
			alt1.accept(this);
		else
			alt2.accept(this);
	}

	protected void visitExceptionType(SyntaxTreeBranch exceptionType) {
		super.visitBranch(exceptionType);
		builder.addThrowRef();
	}
	
	protected void visitStaticInitializer(SyntaxTreeBranch staticInitializer) {
		// savedModifiers = getLeaf(staticInitializer, Symbol.T_STATIC, 0, 0);
		blockIsStatic = true;

		SyntaxTreeBranch block = getBranch(staticInitializer, Symbol.NT_BLOCK, 1, 1);
		block.accept(this);
	}	
	
	protected void visitConstructorDeclaration(SyntaxTreeBranch constructorDeclaration) {

		SyntaxTreeBranch constructorDeclarator = getBranch(constructorDeclaration, Symbol.NT_CONSTRUCTOR_DECLARATOR, 0, 1);
		constructorDeclarator.accept(this);

		SyntaxTreeBranch constructorModifiers = getBranch(constructorDeclaration, Symbol.NT_CONSTRUCTOR_MODIFIERS, 0, 0);
		if (constructorModifiers != null)
			constructorModifiers.accept(this);

		SyntaxTreeBranch throze = getBranch(constructorDeclaration, Symbol.NT_THROWS, 1, 2);
		if (throze != null)
			throze.accept(this);

		SyntaxTreeBranch constructorBody = getBranch(constructorDeclaration, Symbol.NT_CONSTRUCTOR_BODY, 1, 3);
		constructorBody.accept(this);

		builder.endConstructor();
	}

	protected void visitConstructorDeclarator(SyntaxTreeBranch constructorDeclarator) {
		SyntaxTreeBranch simpleTypeName = getBranch(constructorDeclarator, Symbol.NT_SIMPLE_TYPE_NAME, 0, 1);
		// simpleTypeName.accept(this);
		// name must match containing class.

		builder.startConstructor(constructorDeclarator);

		// Get parameters.
		SyntaxTreeBranch formalParameterList = getBranch(constructorDeclarator, Symbol.NT_FORMAL_PARAMETER_LIST, 2, 3);
        if (formalParameterList != null)
            formalParameterList.accept(this);

		// Defer calling endConstructor until after body.
	}

	protected void visitConstructorBody(SyntaxTreeBranch constructorBody) {
		builder.startBlock(constructorBody);
		super.visitBranch(constructorBody);
		builder.endBlock();
	}

	protected void visitExplicitConstructorInvocation(SyntaxTreeBranch explicitConstructorInvocation) {
        SyntaxTreeLeaf thisLeaf = getLeaf(explicitConstructorInvocation, Symbol.T_THIS, 0, 0);
        SyntaxTreeLeaf superLeaf = getLeaf(explicitConstructorInvocation, Symbol.T_SUPER, 0, 0);
        SyntaxTreeBranch primary = getBranch(explicitConstructorInvocation, Symbol.NT_PRIMARY, 0, 0);
        SyntaxTreeBranch expressionName = getBranch(explicitConstructorInvocation, Symbol.NT_EXPRESSION_NAME, 0, 0);

		if (thisLeaf != null)
			builder.lookupThis(thisLeaf);
		else if (superLeaf != null)
			builder.lookupSuper(superLeaf);
		else if (primary != null) {
			primary.accept(this);
			superLeaf = getLeaf(explicitConstructorInvocation, Symbol.T_SUPER, 2, 3);
			builder.lookupOuterSuper(superLeaf);
		}
		else {
			expressionName.accept(this);
			superLeaf = getLeaf(explicitConstructorInvocation, Symbol.T_SUPER, 2, 3);
			builder.lookupOuterSuper(superLeaf);
		}

		builder.lookupConstructor();

        SyntaxTreeBranch nonWildTypeArguments = getBranch(explicitConstructorInvocation, Symbol.NT_NON_WILD_TYPE_ARGUMENTS, 0, 2);
		if (nonWildTypeArguments != null)
			nonWildTypeArguments.accept(this);

        SyntaxTreeBranch argumentList = getBranch(explicitConstructorInvocation, Symbol.NT_ARGUMENT_LIST, 2, 5);
		if (argumentList != null)
			argumentList.accept(this);

		builder.addReference();
	}

	private void visitReferenceTypeList(SyntaxTreeBranch referenceTypeList) {
		builder.startOpTypeArgument();
		super.visitBranch(referenceTypeList);
		builder.endOpTypeArgument();
    }	

	protected void visitEnumDeclaration(SyntaxTreeBranch enumDeclaration) {
		SyntaxTreeLeaf ident = getLeaf(enumDeclaration, Symbol.T_IDENTIFIER, 1, 2);
        builder.startEnum(ident.getValue(), enumDeclaration);

		super.visitBranch(enumDeclaration);
		builder.endEnum();
	}
	
	protected void visitEnumConstant(SyntaxTreeBranch enumConstant) {
		/* ToDo !!!
		 *
	    SyntaxTreeLeaf id = getLeaf(enumConstant, Symbol.T_IDENTIFIER, 0, 1);
 		
		builder.startField(id.getValue(), enumConstant);
		
		builder.addModifier(Modifier.PUBLIC);
		builder.addModifier(Modifier.STATIC);
		builder.addModifier(Modifier.FINAL);
		
		SyntaxTreeBranch parent = enumConstant.getParent();
		while(getType(parent) != Symbol.NT_ENUM_DECLARATION)
			parent = parent.getParent();
		
		SyntaxTreeLeaf enumId = getLeaf(parent, Symbol.T_IDENTIFIER, 1, 2);
		builder.setType(enumId.getValue());
		
		// ToDo!!! Deal with constant-specific class body.
		
		builder.endField();
		*/
	}

	protected void visitNormalInterfaceDeclaration(SyntaxTreeBranch normalInterfaceDeclaration) {
		SyntaxTreeLeaf ident = getLeaf(normalInterfaceDeclaration, Symbol.T_IDENTIFIER, 1, 2);
        builder.startInterface(ident.getValue(), normalInterfaceDeclaration);

		super.visitBranch(normalInterfaceDeclaration);
		builder.endInterface();
	}

	protected void visitExtendsInterfaces(SyntaxTreeBranch extendsInterfaces) {

        SyntaxTreeBranch leftExtendsInterfaces = getBranch(extendsInterfaces, Symbol.NT_EXTENDS_INTERFACES, 0, 0);
		if (leftExtendsInterfaces != null)
			leftExtendsInterfaces.accept(this);

        SyntaxTreeBranch interfaceType = getBranch(extendsInterfaces, Symbol.NT_INTERFACE_TYPE, 0, 2);
		
		typeIsInterface = true;
		super.visitBranch(interfaceType);

		builder.superInterface();
	}

	/*
	protected void visitInterfaceBody(SyntaxTreeBranch interfaceBody) {
		builder.startBlock(interfaceBody);
		super.visitBranch(interfaceBody);
		builder.endBlock();
	}
	*/

	protected void visitConstantDeclaration(SyntaxTreeBranch constantDeclaration) {

		// Don't visit type or modifiers until lower down, when we have actually declared a field.
		savedModifiers = getBranch(constantDeclaration, Symbol.NT_CONSTANT_MODIFIERS,  0, 0);
		savedType = getBranch(constantDeclaration, Symbol.NT_TYPE,  0, 1);

		SyntaxTreeBranch variableDeclarators = getBranch(constantDeclaration, Symbol.NT_VARIABLE_DECLARATORS, 1, 2);

		declaringField = true;
		variableDeclarators.accept(this);
		declaringField = false;
    }	

	protected void visitAbstractMethodDeclaration(SyntaxTreeBranch abstractMethodDeclaration) {

		SyntaxTreeBranch methodDeclarator = getBranch(abstractMethodDeclaration, Symbol.NT_METHOD_DECLARATOR, 1, 3);
		methodDeclarator.accept(this);

		SyntaxTreeBranch abstractMethodModifiers = getBranch(abstractMethodDeclaration, Symbol.NT_ABSTRACT_METHOD_MODIFIERS, 0, 0);
		if (abstractMethodModifiers != null)
			abstractMethodModifiers.accept(this);

		SyntaxTreeBranch typeParameters = getBranch(abstractMethodDeclaration, Symbol.NT_TYPE_PARAMETERS, 0, 1);
		if (typeParameters != null)
			typeParameters.accept(this);

		SyntaxTreeBranch resultType = getBranch(abstractMethodDeclaration, Symbol.NT_RESULT_TYPE, 0, 2);
		resultType.accept(this);

		while (extraDims > 0) {
			builder.lookupArray();
			extraDims--;
		}

		builder.setType();

		SyntaxTreeBranch throze = getBranch(abstractMethodDeclaration, Symbol.NT_THROWS, 2, 4);
		if (throze != null)
			throze.accept(this);

		builder.endMethod();
	}

	protected void visitBlock(SyntaxTreeBranch block) {
		builder.startBlock(block);

		// Static can only apply to outermost block.
		boolean staticInitBlock = blockIsStatic;
		blockIsStatic = false;

		super.visitBranch(block);

		if (staticInitBlock);
			builder.addModifier(Modifier.STATIC);

		builder.endBlock();
	}

	/*
	protected void visitBlockStatement(SyntaxTreeBranch blockStatement) {
		super.visitBranch(blockStatement);

		if (getBranch(blockStatement, Symbol.NT_STATEMENT,  0, 0) != null)
			builder.addReference();
	}
	*/

	// One local variable declaration might actually declare many variables.
	protected void visitLocalVariableDeclaration(SyntaxTreeBranch localVariableDeclaration) {

		// Don't visit type or modifiers until lower down, when we have actually declared a variable.
		savedModifiers = getBranch(localVariableDeclaration, Symbol.NT_VARIABLE_MODIFIERS,  0, 0);
		savedType = getBranch(localVariableDeclaration, Symbol.NT_TYPE,  0, 1);

		SyntaxTreeBranch variableDeclarators = getBranch(localVariableDeclaration, Symbol.NT_VARIABLE_DECLARATORS, 1, 2);

		declaringLocal = true;
		variableDeclarators.accept(this);
		declaringLocal = false;
    }

	protected void visitStatementExpression(SyntaxTreeBranch statementExpression) {
		super.visitBranch(statementExpression);
		builder.addReference();
	}

	protected void visitIfThenStatement(SyntaxTreeBranch ifThenStatement) {
		SyntaxTreeBranch expression = getBranch(ifThenStatement, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statement = getBranch(ifThenStatement, Symbol.NT_STATEMENT, 4, 4);
		statement.accept(this);
	}

	protected void visitIfThenElseStatement(SyntaxTreeBranch ifThenElseStatement) {
		SyntaxTreeBranch expression = getBranch(ifThenElseStatement, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statementNoShortIf = getBranch(ifThenElseStatement, Symbol.NT_STATEMENT_NO_SHORT_IF, 4, 4);
		statementNoShortIf.accept(this);

		SyntaxTreeBranch statement = getBranch(ifThenElseStatement, Symbol.NT_STATEMENT, 6, 6);
		statement.accept(this);
	}

	protected void visitIfThenElseStatementNoShortIf(SyntaxTreeBranch ifThenElseStatementNoShortIf) {
		SyntaxTreeBranch expression = getBranch(ifThenElseStatementNoShortIf, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statementNoShortIf1 = getBranch(ifThenElseStatementNoShortIf, Symbol.NT_STATEMENT_NO_SHORT_IF, 4, 4);
		statementNoShortIf1.accept(this);

		SyntaxTreeBranch statementNoShortIf2 = getBranch(ifThenElseStatementNoShortIf, Symbol.NT_STATEMENT_NO_SHORT_IF, 6, 6);
		statementNoShortIf2.accept(this);
	}

	protected void visitAssertStatement(SyntaxTreeBranch assertStatement) {
		SyntaxTreeBranch expression1 = getBranch(assertStatement, Symbol.NT_EXPRESSION, 1, 1);
		expression1.accept(this);
		builder.addReference();

		SyntaxTreeBranch expression2 = getBranch(assertStatement, Symbol.NT_EXPRESSION, 3, 3);
		if (expression2 != null) {
			expression2.accept(this);
			builder.addReference();
		}
	}

	protected void visitSwitchStatement(SyntaxTreeBranch switchStatement) {
		SyntaxTreeBranch expression = getBranch(switchStatement, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch switchBlock = getBranch(switchStatement, Symbol.NT_SWITCH_BLOCK, 4, 4);
		switchBlock.accept(this);
	}

	protected void visitSwitchLabel(SyntaxTreeBranch switchLabel) {
		//UGLY: if the expression is an identifier, it will be ambiguous.
		//Ignored for now.

		super.visitBranch(switchLabel); // bypass normal visitBlock() method.

		if (getLeaf(switchLabel, Symbol.T_CASE, 0, 0) != null)
			builder.addReference();
	}

	protected void visitWhileStatement(SyntaxTreeBranch whileStatement) {
		SyntaxTreeBranch expression = getBranch(whileStatement, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statement = getBranch(whileStatement, Symbol.NT_STATEMENT, 4, 4);
		statement.accept(this);
	}

	protected void visitWhileStatementNoShortIf(SyntaxTreeBranch whileStatementNoShortIf) {
		SyntaxTreeBranch expression = getBranch(whileStatementNoShortIf, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statementNoShortIf = getBranch(whileStatementNoShortIf, Symbol.NT_STATEMENT_NO_SHORT_IF, 4, 4);
		statementNoShortIf.accept(this);
	}

	protected void visitDoStatement(SyntaxTreeBranch doStatement) {
		SyntaxTreeBranch statement = getBranch(doStatement, Symbol.NT_STATEMENT, 1, 1);
		statement.accept(this);

		SyntaxTreeBranch expression = getBranch(doStatement, Symbol.NT_EXPRESSION, 4, 4);
		expression.accept(this);
		builder.addReference();
	}

	protected void visitBasicForStatement(SyntaxTreeBranch basicForStatement) {
		SyntaxTreeBranch forInit = getBranch(basicForStatement, Symbol.NT_FOR_INIT, 2, 2);
		if (forInit != null)
			forInit.accept(this);

		SyntaxTreeBranch expression = getBranch(basicForStatement, Symbol.NT_EXPRESSION, 3, 4);
		if (expression != null) {
			expression.accept(this);
			builder.addReference();
		}

		SyntaxTreeBranch forUpdate = getBranch(basicForStatement, Symbol.NT_FOR_UPDATE, 4, 6);
		if (forUpdate != null)
			forUpdate.accept(this);

		SyntaxTreeBranch statement = getBranch(basicForStatement, Symbol.NT_STATEMENT, 5, 8);
		statement.accept(this);
	}

	protected void visitForStatementNoShortIf(SyntaxTreeBranch forStatementNoShortIf) {
		SyntaxTreeBranch forInit = getBranch(forStatementNoShortIf, Symbol.NT_FOR_INIT, 2, 2);
		if (forInit != null)
			forInit.accept(this);

		SyntaxTreeBranch expression = getBranch(forStatementNoShortIf, Symbol.NT_EXPRESSION, 3, 4);
		if (expression != null) {
			expression.accept(this);
			builder.addReference();
		}

		SyntaxTreeBranch forUpdate = getBranch(forStatementNoShortIf, Symbol.NT_FOR_UPDATE, 4, 6);
		if (forUpdate != null)
			forUpdate.accept(this);

		SyntaxTreeBranch statementNoShortIf = getBranch(forStatementNoShortIf, Symbol.NT_STATEMENT_NO_SHORT_IF, 5, 8);
		statementNoShortIf.accept(this);
	}

	protected void visitEnhancedForStatement(SyntaxTreeBranch enhancedForStatement) {
		SyntaxTreeLeaf id = getLeaf(enhancedForStatement, Symbol.T_IDENTIFIER, 3, 4);
		builder.startLocalVariable(id.getValue(), enhancedForStatement);

		SyntaxTreeBranch type = getBranch(enhancedForStatement, Symbol.NT_TYPE, 2, 3);
		type.accept(this);

		SyntaxTreeBranch variableModifiers = getBranch(enhancedForStatement, Symbol.NT_VARIABLE_MODIFIERS, 2, 2);
		if (variableModifiers != null)
			variableModifiers.accept(this);

		builder.endLocalVariable();

		SyntaxTreeBranch expression = getBranch(enhancedForStatement, Symbol.NT_EXPRESSION, 5, 6);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch statement = getBranch(enhancedForStatement, Symbol.NT_STATEMENT, 7, 8);
		statement.accept(this);
	}

	protected void visitReturnStatement(SyntaxTreeBranch returnStatement) {
		SyntaxTreeBranch expression = getBranch(returnStatement, Symbol.NT_EXPRESSION, 1, 1);
		if (expression != null) {
			expression.accept(this);
			builder.addReference();
		}
	}

	protected void visitThrowStatement(SyntaxTreeBranch throwStatement) {
		SyntaxTreeBranch expression = getBranch(throwStatement, Symbol.NT_EXPRESSION, 1, 1);
		expression.accept(this);
		builder.addReference();
	}

	protected void visitSynchronizedStatement(SyntaxTreeBranch synchronizedStatement) {
		SyntaxTreeBranch expression = getBranch(synchronizedStatement, Symbol.NT_EXPRESSION, 2, 2);
		expression.accept(this);
		builder.addReference();

		SyntaxTreeBranch block = getBranch(synchronizedStatement, Symbol.NT_BLOCK, 4, 4);
		block.accept(this);
	}

	protected void visitCatchClause(SyntaxTreeBranch catchClause) {
		builder.startCatchBlock(catchClause);
		
		SyntaxTreeBranch formalParameter = getBranch(catchClause, Symbol.NT_FORMAL_PARAMETER, 2, 2);
		formalParameter.accept(this);
		
		SyntaxTreeBranch block = getBranch(catchClause, Symbol.NT_BLOCK, 4, 4);
		super.visitBranch(block); // bypass normal visitBlock() method.
		
		builder.endCatchBlock();
	}
	
	protected void visitFinally(SyntaxTreeBranch finallee) {
		builder.startCatchBlock(finallee);
		
		super.visitBranch(finallee); // bypass normal visitBlock() method.
		
		builder.endCatchBlock();
	}

	private void visitPrimaryNoNewArray(SyntaxTreeBranch primaryNoNewArray) {
		System.out.println(primaryNoNewArray);
        SyntaxTreeLeaf literal = getLeaf(primaryNoNewArray, Symbol.T_LITERAL, 0, 0);
		if (literal != null) {
			builder.literal(literal);
			return;
		}

		// type.class, void.class
        SyntaxTreeLeaf classLeaf = getLeaf(primaryNoNewArray, Symbol.T_CLASS, 2, 2);
		if (classLeaf != null) {
			super.visitBranch(primaryNoNewArray);
			builder.addReference();
			builder.dotClass(classLeaf);
            return;
        }

		// this, <ClassName>.this
        SyntaxTreeLeaf thisLeaf = getLeaf(primaryNoNewArray, Symbol.T_THIS, 0, 0);
		if (thisLeaf != null) {
			builder.lookupThis(thisLeaf);
			return;
		}

        thisLeaf = getLeaf(primaryNoNewArray, Symbol.T_THIS, 2, 2);
		if (thisLeaf != null) {
			super.visitBranch(primaryNoNewArray);
			builder.lookupOuterThis(thisLeaf);
            return;
        }

		// Special case for <PrimaryNoNewArray> ::= "(" <Expression> ")"
		// We usually throw away the type of an expression (bt calling builder.addReference().
		// We want to keep the expression type here, so skip the usual visitExpression().
        SyntaxTreeBranch expression = getBranch(primaryNoNewArray, Symbol.NT_EXPRESSION, 1, 1);
		if (expression != null) {
			super.visitBranch(expression);
			return;
		}

		// Just walk kids.
        super.visitBranch(primaryNoNewArray);
	}

	// Need to deal with anonymous inner classes.
	protected void visitClassInstanceCreationExpression(SyntaxTreeBranch classInstanceCreationExpression) {
		SyntaxTreeBranch classOrInterfaceType = getBranch(classInstanceCreationExpression, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 1, 2);
		SyntaxTreeBranch primary = getBranch(classInstanceCreationExpression, Symbol.NT_PRIMARY, 0, 0);
		SyntaxTreeBranch expressionName = getBranch(classInstanceCreationExpression, Symbol.NT_EXPRESSION_NAME, 0, 0);

		SyntaxTreeBranch constructorTypeArguments = getBranch(classInstanceCreationExpression, Symbol.NT_TYPE_ARGUMENTS, 0, 2);
		SyntaxTreeLeaf id = getLeaf(classInstanceCreationExpression, Symbol.T_IDENTIFIER, 3, 4);
		SyntaxTreeBranch typeArguments = getBranch(classInstanceCreationExpression, Symbol.NT_TYPE_ARGUMENTS, 4, 5);
		SyntaxTreeBranch argList = getBranch(classInstanceCreationExpression, Symbol.NT_ARGUMENT_LIST, 3, 7);

        SyntaxTreeBranch classBody = getBranch(classInstanceCreationExpression, Symbol.NT_CLASS_BODY, 4, 9);

		if (classBody != null) {
			// Need to declare an anonymous class.
	        builder.startClass(null, classInstanceCreationExpression);
	        
			String abominableName = builder.getDeclName();

	        if (classOrInterfaceType != null) {
				if (argList == null)
					typeIsUserType = true; // Interface is legal too.
				else
					typeIsClass = true; // If we have args, then type must be a ClassType because Interfaces don't have constructors.

	        	classOrInterfaceType.accept(this);
			}
	        else {
				if (primary != null)
					primary.accept(this);
				else
					expressionName.accept(this);

				if (argList == null)
					builder.lookupUserType(id);
				else
					builder.lookupClassType(id);

				if (typeArguments != null)
					typeArguments.accept(this);
			}

			builder.maybeSuper();

	        /* ToDo generate appropriate constructor when base class is static */

            // Anonymous class needs a constructor with the same parameters as the invoked superclass constructor.
			// #15.9.5.1: An anonymous class cannot have an explicitly declared constructor.
            builder.startConstructor(null); // No source.

			boolean wasDeclaringArgs = declaringAnonConstructorArgs; // save prior state

            if (argList != null) {
                // If we have args, then type must be a ClassType because Interfaces don't have constructors.
				declaringAnonConstructorArgs = true;
            	argList.accept(this);
				declaringAnonConstructorArgs = false;
            }
            builder.endConstructor();
	        
			classBody.accept(this);
			
			builder.endClass();

			builder.lookupClassType(abominableName);	// Look up the class we just generated...
			builder.lookupConstructor();				// ... so we can find the constructor.

			if (constructorTypeArguments != null)
				constructorTypeArguments.accept(this);

			if (argList != null)
				argList.accept(this); // 2nd time thru, but flag causes lookup instead of decl

			declaringAnonConstructorArgs = wasDeclaringArgs; // restore prior state.

			return;
		}

		if (classOrInterfaceType != null) {
			typeIsClass = true;
			classOrInterfaceType.accept(this);
		}
		else {
			if (primary != null)
				primary.accept(this);
			else
				expressionName.accept(this);

			builder.lookupClassType(id);

			if (typeArguments != null)
				typeArguments.accept(this);
		}
		
		builder.lookupConstructor();

		if (constructorTypeArguments != null)
			constructorTypeArguments.accept(this);

		if (argList != null)
			argList.accept(this);
	}

	private void visitArgumentList(SyntaxTreeBranch argumentList) {
        SyntaxTreeBranch leftArgumentList = getBranch(argumentList, Symbol.NT_ARGUMENT_LIST, 0, 0);
		if (leftArgumentList != null)
			leftArgumentList.accept(this);

		SyntaxTreeBranch expression = getBranch(argumentList, Symbol.NT_EXPRESSION, 0, 2);
		
		boolean prevDeclaringAnonConstructorArgs = declaringAnonConstructorArgs;
		declaringAnonConstructorArgs = false;
		
		if (prevDeclaringAnonConstructorArgs)
			builder.startParameter(null, expression);
		else
			builder.startArg();

		expression.accept(this);

		if (prevDeclaringAnonConstructorArgs) {
			builder.setType();
			builder.endParameter();
		}
		else
			builder.endArg();
		
		declaringAnonConstructorArgs = prevDeclaringAnonConstructorArgs;
    }

	private void visitArrayCreationExpression(SyntaxTreeBranch arrayCreationExpression) {
		// Do initializer first, so type of array remains.
    	SyntaxTreeBranch arrayInitializer = getBranch(arrayCreationExpression, Symbol.NT_ARRAY_INITIALIZER, 3, 3);

        if (arrayInitializer != null)
            arrayInitializer.accept(this); // May eventually contain expressions.

    	SyntaxTreeBranch typeNode = getBranch(arrayCreationExpression, Symbol.NT_PRIMITIVE_TYPE, 1, 1);
		if (typeNode == null)
	    	typeNode = getBranch(arrayCreationExpression, Symbol.NT_CLASS_OR_INTERFACE_TYPE, 1, 1);

		typeNode.accept(this);

        SyntaxTreeBranch dimExprs = getBranch(arrayCreationExpression, Symbol.NT_DIM_EXPRS, 2, 2);
		SyntaxTreeBranch dims = getBranch(arrayCreationExpression, Symbol.NT_DIMS, 2, 3);
        if (dimExprs != null)
            dimExprs.accept(this);

		if (dims != null)
			dims.accept(this);
	}

	private void visitDimExpr(SyntaxTreeBranch dimExpr) {
		builder.start2ndExpression();
		super.visitBranch(dimExpr);
		builder.end2ndExpression();

		builder.lookupArray();
	}

	private void visitDims(SyntaxTreeBranch dims) {
		super.visitBranch(dims);
		builder.lookupArray();
	}

	private void visitFieldAccess(SyntaxTreeBranch fieldAccess) {
        SyntaxTreeBranch primary = getBranch(fieldAccess, Symbol.NT_PRIMARY, 0, 0);
        SyntaxTreeLeaf superLeaf = getLeaf(fieldAccess, Symbol.T_SUPER, 0, 0);
		SyntaxTreeBranch className = getBranch(fieldAccess, Symbol.NT_CLASS_NAME, 0, 0);

		if (primary != null)
			primary.accept(this);

		else if (superLeaf != null) // super.foo
			builder.lookupSuper(superLeaf);

		else {
			className.accept(this);
			superLeaf = getLeaf(fieldAccess, Symbol.T_SUPER, 2, 2);
			builder.lookupOuterSuper(superLeaf);
		}

		SyntaxTreeLeaf id = getLeaf(fieldAccess, Symbol.T_IDENTIFIER, 2,  4);
		builder.lookupVariable(id);
    }

	private void visitMethodInvocation(SyntaxTreeBranch methodInvocation) {

        SyntaxTreeBranch methodName = getBranch(methodInvocation, Symbol.NT_METHOD_NAME, 0, 0);

        if (methodName != null) {
			super.visitBranch(methodInvocation);
			return;
        }

        SyntaxTreeBranch primary = getBranch(methodInvocation, Symbol.NT_PRIMARY, 0, 0);
        SyntaxTreeLeaf superLeaf = getLeaf(methodInvocation, Symbol.T_SUPER, 0, 0);
		SyntaxTreeBranch className = getBranch(methodInvocation, Symbol.NT_CLASS_NAME, 0, 0);
		SyntaxTreeBranch typeName = getBranch(methodInvocation, Symbol.NT_TYPE_NAME, 0, 0);

		if (primary != null)
			primary.accept(this);

		else if (superLeaf != null) // super.foo
			builder.lookupSuper(superLeaf);

		else if (className != null) {
			className.accept(this);
			superLeaf = getLeaf(methodInvocation, Symbol.T_SUPER, 2, 2);
			builder.lookupOuterSuper(superLeaf);
		}
		else
			typeName.accept(this);

		SyntaxTreeLeaf id = getLeaf(methodInvocation, Symbol.T_IDENTIFIER, 2,  5);
		builder.lookupMethod(id);

        SyntaxTreeLeaf nonWildTypeArguments = getLeaf(methodInvocation, Symbol.NT_NON_WILD_TYPE_ARGUMENTS, 2, 4);
		if (nonWildTypeArguments != null)
			nonWildTypeArguments.accept(this);

		SyntaxTreeBranch argumentList =  getBranch(methodInvocation, Symbol.NT_ARGUMENT_LIST, 4, 7);
		if (argumentList != null)
			argumentList.accept(this);
	}

	private void visitArrayAccess(SyntaxTreeBranch arrayAccess) {
    	// Make sure we walk through the index expression, even though we don't use its type.
    	SyntaxTreeBranch expression = getBranch(arrayAccess, Symbol.NT_EXPRESSION, 2, 2);
        expression.accept(this);
		builder.addReference();

        SyntaxTreeBranch expressionName = getBranch(arrayAccess, Symbol.NT_EXPRESSION_NAME, 0, 0);
        SyntaxTreeBranch primaryNoNewArray = getBranch(arrayAccess, Symbol.NT_PRIMARY_NO_NEW_ARRAY, 0, 0);

		if (expressionName != null)
			expressionName.accept(this);
		else
			primaryNoNewArray.accept(this);

		builder.addArrayAccess();
	}

	private void visitUnaryExpressionNotPlusMinus(SyntaxTreeBranch unaryExpressionNotPlusMinus) {
		super.visitBranch(unaryExpressionNotPlusMinus);

		if (getLeaf(unaryExpressionNotPlusMinus, Symbol.T_COMP, 0,  0) != null)
			builder.unaryPromotion();
	}

	private void visitCastExpression(SyntaxTreeBranch castExpression) {
        SyntaxTreeBranch valueExp = getBranch(castExpression, Symbol.NT_UNARY_EXPRESSION, 3, 4);
		if (valueExp == null)
			valueExp = getBranch(castExpression, Symbol.NT_UNARY_EXPRESSION_NOT_PLUS_MINUS, 3, 3);

		valueExp.accept(this);
		builder.addReference();

        SyntaxTreeBranch type = getBranch(castExpression, Symbol.NT_PRIMITIVE_TYPE, 1, 1);
		if (type == null)
			type = getBranch(castExpression, Symbol.NT_REFERENCE_TYPE, 1, 1);

		type.accept(this);

        SyntaxTreeBranch dims = getBranch(castExpression, Symbol.NT_DIMS, 2, 2);
		if (dims != null)
			dims.accept(this);

		// builder.isInstance(); Damn. gotta work on arrays & primitives too.
	}

	private void visitMultiplicativeExpression(SyntaxTreeBranch multiplicativeExpression) {
        SyntaxTreeBranch unaryExpression = getBranch(multiplicativeExpression, Symbol.NT_UNARY_EXPRESSION, 0, 0);
    	if (unaryExpression != null) {
            unaryExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsMultiplicativeExpression = getBranch(multiplicativeExpression, Symbol.NT_MULTIPLICATIVE_EXPRESSION, 0, 0);
		lhsMultiplicativeExpression.accept(this);

		builder.start2ndExpression();
		SyntaxTreeBranch rhsUnaryExpression = getBranch(multiplicativeExpression, Symbol.NT_UNARY_EXPRESSION, 2, 2);
        rhsUnaryExpression.accept(this);

		builder.binaryPromotion();
	}

	private void visitAdditiveExpression(SyntaxTreeBranch additiveExpression) {
		
        SyntaxTreeBranch multiplicativeExpression = getBranch(additiveExpression, Symbol.NT_MULTIPLICATIVE_EXPRESSION, 0, 0);
    	if (multiplicativeExpression != null) {
            multiplicativeExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsAdditiveExpression = getBranch(additiveExpression, Symbol.NT_ADDITIVE_EXPRESSION, 0, 0);
		lhsAdditiveExpression.accept(this);

		builder.start2ndExpression();
		SyntaxTreeBranch rhsMultiplicativeExpression = getBranch(additiveExpression, Symbol.NT_MULTIPLICATIVE_EXPRESSION, 2, 2);
		rhsMultiplicativeExpression.accept(this);

		builder.binaryPromotionOrStringConcatenation(); // Bleuch.
	}

	private void visitShiftExpression(SyntaxTreeBranch shiftExpression) {
        SyntaxTreeBranch additiveExpression = getBranch(shiftExpression, Symbol.NT_ADDITIVE_EXPRESSION, 0, 0);
    	if (additiveExpression != null) {
            additiveExpression.accept(this);
            return;
        }

		// Do rhs first, because type comes from the lhs.
		SyntaxTreeBranch rhsAdditiveExpression = getBranch(shiftExpression, Symbol.NT_ADDITIVE_EXPRESSION, 2, 2);
		if(rhsAdditiveExpression != null) {
			rhsAdditiveExpression.accept(this); // Not interested in type.
			builder.addReference();
		}

		SyntaxTreeBranch lhsShiftExpression = getBranch(shiftExpression, Symbol.NT_SHIFT_EXPRESSION, 0, 0);
		lhsShiftExpression.accept(this);
		builder.unaryPromotion();
	}

	private void visitRelationalExpression(SyntaxTreeBranch relationalExpression) {
        SyntaxTreeBranch shiftExpression = getBranch(relationalExpression, Symbol.NT_SHIFT_EXPRESSION, 0, 0);
    	if (shiftExpression != null) {
            shiftExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsRelationalExpression = getBranch(relationalExpression, Symbol.NT_RELATIONAL_EXPRESSION, 0, 0);
		lhsRelationalExpression.accept(this);
		builder.addReference();

		SyntaxTreeBranch rhsShiftExpression = getBranch(relationalExpression, Symbol.NT_SHIFT_EXPRESSION, 2, 2);
		SyntaxTreeBranch referenceType = getBranch(relationalExpression, Symbol.NT_REFERENCE_TYPE, 2, 2);

        if (rhsShiftExpression != null)
			rhsShiftExpression.accept(this);
        else
			referenceType.accept(this);

		builder.addReference();
		builder.booleanType();
	}

	private void visitEqualityExpression(SyntaxTreeBranch equalityExpression) {
        SyntaxTreeBranch relationalExpression = getBranch(equalityExpression, Symbol.NT_RELATIONAL_EXPRESSION, 0, 0);
    	if (relationalExpression != null) {
            relationalExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsEqualityExpression = getBranch(equalityExpression, Symbol.NT_EQUALITY_EXPRESSION, 0, 0);
		lhsEqualityExpression.accept(this);
		builder.addReference();

		SyntaxTreeBranch rhsRelationalExpression = getBranch(equalityExpression, Symbol.NT_RELATIONAL_EXPRESSION, 2, 2);
		rhsRelationalExpression.accept(this);
		builder.addReference();

		builder.booleanType();
	}

	private void visitAndExpression(SyntaxTreeBranch andExpression) {
        SyntaxTreeBranch equalityExpression = getBranch(andExpression, Symbol.NT_EQUALITY_EXPRESSION, 0, 0);
    	if (equalityExpression != null) {
            equalityExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsAndExpression = getBranch(andExpression, Symbol.NT_AND_EXPRESSION, 0, 0);
		lhsAndExpression.accept(this);
		builder.start2ndExpression();

		SyntaxTreeBranch rhsEqualityExpression = getBranch(andExpression, Symbol.NT_EQUALITY_EXPRESSION, 2, 2);
		rhsEqualityExpression.accept(this);

		builder.binaryPromotionOrBoolean(); // Bleuch.
	}

	private void visitExclusiveOrExpression(SyntaxTreeBranch exclusiveOrExpression) {
        SyntaxTreeBranch andExpression = getBranch(exclusiveOrExpression, Symbol.NT_AND_EXPRESSION, 0, 0);
    	if (andExpression != null) {
            andExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsExclusiveOrExpression = getBranch(exclusiveOrExpression, Symbol.NT_EXCLUSIVE_OR_EXPRESSION, 0, 0);
		lhsExclusiveOrExpression.accept(this);
		builder.start2ndExpression();

		SyntaxTreeBranch rhsAndExpression = getBranch(exclusiveOrExpression, Symbol.NT_AND_EXPRESSION, 2, 2);
		rhsAndExpression.accept(this);

		builder.binaryPromotionOrBoolean(); // Bleuch.
	}

	private void visitInclusiveOrExpression(SyntaxTreeBranch inclusiveOrExpression) {
        SyntaxTreeBranch exclusiveOrExpression = getBranch(inclusiveOrExpression, Symbol.NT_EXCLUSIVE_OR_EXPRESSION, 0, 0);
    	if (exclusiveOrExpression != null) {
            exclusiveOrExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsInclusiveOrExpression = getBranch(inclusiveOrExpression, Symbol.NT_INCLUSIVE_OR_EXPRESSION, 0, 0);
		lhsInclusiveOrExpression.accept(this);
		builder.start2ndExpression();

		SyntaxTreeBranch rhsExclusiveOrExpression = getBranch(inclusiveOrExpression, Symbol.NT_EXCLUSIVE_OR_EXPRESSION, 2, 2);
		rhsExclusiveOrExpression.accept(this);

		builder.binaryPromotionOrBoolean(); // Bleuch.
	}

	private void visitConditionalAndExpression(SyntaxTreeBranch conditionalAndExpression) {
        SyntaxTreeBranch inclusiveOrExpression = getBranch(conditionalAndExpression, Symbol.NT_INCLUSIVE_OR_EXPRESSION, 0, 0);
    	if (inclusiveOrExpression != null) {
            inclusiveOrExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsConditionalAndExpression = getBranch(conditionalAndExpression, Symbol.NT_CONDITIONAL_AND_EXPRESSION, 0, 0);
		lhsConditionalAndExpression.accept(this);
		builder.addReference();

    	SyntaxTreeBranch rhsInclusiveOrExpression = getBranch(conditionalAndExpression, Symbol.NT_INCLUSIVE_OR_EXPRESSION, 2, 2);
		rhsInclusiveOrExpression.accept(this);
       	// Leave rhs result as current.
	}

	private void visitConditionalOrExpression(SyntaxTreeBranch conditionalOrExpression) {
        SyntaxTreeBranch conditionalAndExpression = getBranch(conditionalOrExpression, Symbol.NT_CONDITIONAL_AND_EXPRESSION, 0, 0);
    	if (conditionalAndExpression != null) {
            conditionalAndExpression.accept(this);
            return;
        }

		SyntaxTreeBranch lhsConditionalOrExpression = getBranch(conditionalOrExpression, Symbol.NT_CONDITIONAL_OR_EXPRESSION, 0, 0);
		lhsConditionalOrExpression.accept(this);
		builder.addReference();

		SyntaxTreeBranch rhsConditionalAndExpression = getBranch(conditionalOrExpression, Symbol.NT_CONDITIONAL_AND_EXPRESSION, 2, 2);
		rhsConditionalAndExpression.accept(this);
       	// Leave rhs result as current.
	}

	private void visitConditionalExpression(SyntaxTreeBranch conditionalExpression) {
    	SyntaxTreeBranch conditionalOrExpression = getBranch(conditionalExpression, Symbol.NT_CONDITIONAL_OR_EXPRESSION, 0, 0);
        conditionalOrExpression.accept(this);

        SyntaxTreeBranch expression = getBranch(conditionalExpression, Symbol.NT_EXPRESSION, 2, 2);
        if (expression == null)
        	return;

		// OrExpr ? expr : condExpr

		builder.addReference(); // Record OrExpr type.

		expression.accept(this);
		builder.start2ndExpression();

    	SyntaxTreeBranch innerConditionalExpression = getBranch(conditionalExpression, Symbol.NT_CONDITIONAL_EXPRESSION, 4, 4);
		innerConditionalExpression.accept(this);
		builder.conditionalExpression();
	}

	private void visitAssignment(SyntaxTreeBranch assignment) {
    	// lhs op expr

    	SyntaxTreeBranch assignmentExpression = getBranch(assignment, Symbol.NT_ASSIGNMENT_EXPRESSION, 2, 2);
        assignmentExpression.accept(this);
		builder.addReference();

    	SyntaxTreeBranch leftHandSide = getBranch(assignment, Symbol.NT_LEFT_HAND_SIDE, 0, 0);
        leftHandSide.accept(this);
	}

	protected void visitExpression(SyntaxTreeBranch expression) {
		super.visitBranch(expression);
		// builder.addReference();
	}

    public void visitLeaf(SyntaxTreeLeaf leaf) {
        switch (getType(leaf)) {
        	case Symbol.T_PUBLIC       : builder.addModifier(Modifier.PUBLIC); return;
        	case Symbol.T_PROTECTED    : builder.addModifier(Modifier.PROTECTED); return;
        	case Symbol.T_PRIVATE      : builder.addModifier(Modifier.PRIVATE); return;
        	case Symbol.T_STATIC       : builder.addModifier(Modifier.STATIC); return;
        	case Symbol.T_ABSTRACT     : builder.addModifier(Modifier.ABSTRACT); return;
        	case Symbol.T_FINAL        : builder.addModifier(Modifier.FINAL); return;
        	case Symbol.T_NATIVE       : builder.addModifier(Modifier.NATIVE); return;
        	case Symbol.T_SYNCHRONIZED : builder.addModifier(Modifier.SYNCHRONIZED); return;
        	case Symbol.T_TRANSIENT    : builder.addModifier(Modifier.TRANSIENT); return;
        	case Symbol.T_VOLATILE     : builder.addModifier(Modifier.VOLATILE); return;
        	case Symbol.T_STRICTFP     : builder.addModifier(Modifier.STRICT); return;

        	case Symbol.T_VOID			: builder.voidType(leaf); return;
        	case Symbol.T_BOOLEAN		: builder.booleanType(leaf); return;
        	case Symbol.T_BYTE			: builder.byteType(leaf); return;
        	case Symbol.T_SHORT			: builder.shortType(leaf); return;
        	case Symbol.T_INT			: builder.intType(leaf); return;
        	case Symbol.T_LONG			: builder.longType(leaf); return;
        	case Symbol.T_CHAR			: builder.charType(leaf); return;
        	case Symbol.T_FLOAT			: builder.floatType(leaf); return;
        	case Symbol.T_DOUBLE		: builder.doubleType(leaf); return;

			default						: super.visitLeaf(leaf);
        }
    }

}
