
package jst.symtab;
import jst.app.JSTMessage;
import jst.parsetree.*;
import jst.symtab.link.*;

import java.util.*;
import java.io.*;
import java.lang.reflect.Modifier;

import jst.symtab.visitor.ModelVisitor;

public abstract class InterfaceType extends UserType {

    /** 
     * @supplierCardinality 0..*
     * @supplierRole superInterfaces*/
	protected Set<InterfaceTypeLookup> superInterfaceRefs;

    /** 
     * @supplierCardinality 0..*
     * @supplierRole subInterfaces*/
	protected Set<InterfaceType> subInterfaces;
	
	//added to support reverse relationships
	protected Set<ClassType> implementedBy;

	protected InterfaceType(SourceFile ownerSourceFile, String theSimpleName, JstParseTreeBranch theSource) {
    	super(ownerSourceFile, theSimpleName, theSource);
    }

	protected InterfaceType(PackageDecl ownerPackage, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerPackage, theSimpleName, theSource);
    }

	protected InterfaceType(UserType ownerUserType, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerUserType, theSimpleName, theSource);
    }

	protected InterfaceType(Block ownerBlock, String theSimpleName, JstParseTreeBranch theSource) {
		super(ownerBlock, theSimpleName, theSource);
    }

	private InterfaceType(GenericInterfaceType theOwner, String theSimpleName, JstParseTreeBranch theSource) {
    	super(theOwner, theSimpleName, theSource);
    }

	protected InterfaceType(SourceFile newOwnerSourceFile, InterfaceType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerSourceFile, that, swaps);
	}

	protected InterfaceType(PackageDecl newOwnerPackage, InterfaceType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerPackage, that, swaps);
	}

	protected InterfaceType(UserType newOwnerUserType, InterfaceType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerUserType, that, swaps);
	}

	protected InterfaceType(Block newOwnerBlock, InterfaceType that, Map<TypeParameter, ActualType> swaps) {
		super(newOwnerBlock, that, swaps);
	}

	protected InterfaceType(GenericInterfaceType newOwner,  InterfaceType that, Map<TypeParameter, ActualType> swaps) {
    	super(newOwner, that, swaps);
	}


    protected void init() {
		super.init();

        superInterfaceRefs = new HashSet<InterfaceTypeLookup>();
        subInterfaces = new HashSet<InterfaceType>();
        implementedBy = new HashSet<ClassType>();
	}

	protected void init(InterfaceType that, Map<TypeParameter, ActualType> swaps) {
		for (InterfaceTypeLookup thatFace : that.superInterfaceRefs)
			addSuperInterfaceRef(thatFace.copy(this));
	}

	public InterfaceType asInterface() {
		return this;
	}

	public void addSuperInterfaceRef(InterfaceTypeLookup interfaceType) {
    	superInterfaceRefs.add(interfaceType);
    }

	public boolean resolveSuperInterfaces(List messages) {

		boolean ok = true;

    	for(InterfaceTypeLookup interfaceRef : superInterfaceRefs){

			if (!interfaceRef.resolve()) {
				ok = false;
				messages.add(interfaceRef.getMessage());
			}
			else
				interfaceRef.getInterfaceType().addSubInterface(this);
		}

		return ok;
	}

	public Set<InterfaceType> getInterfaces() {
		return getSuperInterfaces();
	}

    public Set<InterfaceType> getSuperInterfaces() {
    	Set<InterfaceType> superInterfaces = new HashSet<InterfaceType>();

    	for(InterfaceTypeLookup ref : superInterfaceRefs)
			if (ref.getInterfaceType() != null)
				superInterfaces.add(ref.getInterfaceType());

        return superInterfaces;
    }

    protected void addSubInterface(InterfaceType iface) {
        subInterfaces.add(iface);
    }
    
    public Set getSubInterfaces() {
        return subInterfaces;
    }
	
	public void addImplementedBy(ClassType clazz) {
		implementedBy.add(clazz);
	}
	
	public Set getImplementedBy() {
		return implementedBy;
	}

	public boolean isSubtypeOf(ReferenceType parent) {
		if (parent == this)
        	return true;

		Iterator ifaceIter = getSuperInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			if (iface.isSubtypeOf(parent))
            	return true;
        }

		return false;
    }

	// Simple names only
	public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
    	// Tricky: The name of an interface needs to appear in its own scope.  This is necessary
        // in cases such as import statements, where the import must bring the interface name into scope.
        // However, we don't want the interface to contain itself, so just check for a lookup of the
        // interface name.
        if (typeName.equals(simpleName))
        	return this;

    	TypeDecl result = (UserType) types.get(typeName);
        if (result != null)
			return result.accessible(looker)? result : null;

		Iterator ifaceIter = getSuperInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			result = iface.lookupType(false, typeName, looker);
			if (result != null)
				return result.accessible(looker)? result : null;
		}

		return widen? getOwnerScope().lookupType(true, typeName, looker) : null;
   }

	protected Set findInheritedMethodsNamed(String methodName) {
    	Set result = new HashSet();

		Iterator ifaceIter = getSuperInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			result.addAll(iface.findAllMethodsNamed(methodName));
        }

        // Interfaces appear to inherit the methods of Object!
        if (getSuperInterfaces().isEmpty()) // Root interface
            result.addAll(objct.findAllMethodsNamed(methodName));

		return result;
    }

	public VariableDecl lookupVariable(boolean widen, String variableName, ReferenceType looker) {
    	FieldDecl field = (FieldDecl) fields.get(variableName);

        if (field != null)
			return field.accessible(looker)? field : null;

		Iterator ifaceIter = getSuperInterfaces().iterator();
		while (ifaceIter.hasNext()) {
			InterfaceType iface = (InterfaceType) ifaceIter.next();
			VariableDecl variable = iface.lookupVariable(false, variableName, looker);
			if (variable != null)
				return variable;
		}

        return widen? getOwnerScope().lookupVariable(true, variableName, looker) : null;
	}

    public void accept(ModelVisitor visitor) {
        visitor.visitInterfaceType(this);
    }
}
