
package jst.symtab;

import java.util.*;
import java.io.*;

import jst.app.JSTMessage;
import jst.parsetree.*;
import jst.symtab.link.*;
import jst.symtab.visitor.ModelVisitor;

public class SourceFile extends Decl implements Scope {

	// Derive ".java" source file name from ".xml" parse tree file.
	static protected String getSourceFileName(ParseTree parseTree) {
        String name = parseTree.getFileName(); // No directory prefix.
        int dot = name.lastIndexOf('.');
        return name.substring(0, dot) + ".java";
	}

	protected ParseTree parseTree;

    /**
     * @supplierRole PackageOrTypeLookup
     * @supplierCardinality 0..* */
    private Set<PackageOrTypeLookup> importsOnDemand;

    /**
     * @supplierRole importTypes
     * @supplierCardinality 0..*
     * */
    private Set<UserTypeLookup> importTypes; // UserTypes

    /**
     * @associates <{UserType}>
     * @supplierCardinality 0..*
     * @supplierRole types
     */
    private Map<String, UserType> types;

	/**
     * Construct a source file.
     * At the point of construction, we don't know which package owns the source file (because
     * the package statement is inside the file).  The setPackage() method MUST be called later.
     */
	public SourceFile(ParseTree theParseTree) {
		super(null, getSourceFileName(theParseTree), theParseTree.getRoot());

    	parseTree = theParseTree;
		// pkg = PackageDecl.getDefaultPackage();
        importsOnDemand = new HashSet<PackageOrTypeLookup>();
        importTypes = new HashSet<UserTypeLookup>();
		types = new HashMap<String, UserType>();
    }
	
 	public String getCanonicalName() {
		return getSimpleName();
    }

	// Generate artificial name for anonymous decl in this scope.
    // Sourcefiles do not contain anonymous members.
	public String generateName() {
    	throw new RuntimeException("Source file should not contain anonymous declarations");
    }

	public Scope asScope() {
		return this;
	}

	public SourceFile getOwnerSourceFile() {
		return this;
	}

	public ParseTree getParseTree() {
    	return parseTree;
    }
	
	public void setPackage(PackageDecl pkg) {
    	ownerDecl = pkg;
        pkg.addSourceFile(this);
    }

	public void addImportOnDemandRef(boolean isStatic, PackageOrTypeLookup packageOrTypeLookup) {
		importsOnDemand.add(packageOrTypeLookup);
		if (isStatic)
			System.err.println("ToDo: Add static imports.");
	}

	public boolean resolveImportOnDemands(List<JSTMessage> messages) {

		boolean ok = true;

		Iterator<PackageOrTypeLookup> importIter = importsOnDemand.iterator();
		while (importIter.hasNext()) {
			PackageOrTypeLookup ref = importIter.next();

			if (!ref.resolve()) {
				messages.add(ref.getMessage());
				ok = false;
			}

			if (ref.getUserType() != null)
				ref.getUserType().addImportedOnDemandBy(this);
			else if (ref.getPackageDecl() != null)
				ref.getPackageDecl().addImportedOnDemandBy(this);
		}

		return ok;
	}

	public Set<UserType> getImportOnDemandTypes() {
		Set imports = new HashSet<UserType>();

		Iterator<PackageOrTypeLookup> importIter = importsOnDemand.iterator();
		while (importIter.hasNext()) {
			PackageOrTypeLookup ref = importIter.next();

			UserType type = ref.getUserType();
			if (type != null)
				imports.add(type);
		}

		return imports;
    }

	public Set<PackageDecl> getImportOnDemandPackages() {
		Set imports = new HashSet<PackageDecl>();

		Iterator<PackageOrTypeLookup> importIter = importsOnDemand.iterator();
		while (importIter.hasNext()) {
			PackageOrTypeLookup ref = importIter.next();

			PackageDecl pkg = ref.getPackageDecl();
			if (pkg != null)
				imports.add(pkg);
		}

		return imports;
    }

	public void addImportTypeRef(boolean isStatic, UserTypeLookup userTypeLookup) {
		importTypes.add(userTypeLookup);
		if (isStatic)
			System.err.println("ToDo: Add static imports.");
	}

	public boolean resolveImportTypes(List<JSTMessage> messages) {
		boolean ok = true;

		Iterator<UserTypeLookup> importIter = importTypes.iterator();
		while (importIter.hasNext()) {
			UserTypeLookup ref = importIter.next();

			if (!ref.resolve()) {
				messages.add(ref.getMessage());
				ok = false;
			}
			else {
				UserType userType = ref.getUserType();
				userType.addTypeImportedBy(this);
			}
		}

		return ok;
	}

	public boolean resolveImports(List<JSTMessage> messages) {
		return resolveImportOnDemands(messages) && resolveImportTypes(messages);
	}

	public Set<UserType> getImportTypes() {
		Set<UserType> imports = new HashSet<UserType>();

		Iterator<UserTypeLookup> importIter = importTypes.iterator();
		while (importIter.hasNext()) {
			UserTypeLookup ref = importIter.next();

			UserType userType = ref.getUserType();
			if (userType != null)
				imports.add(userType);
		}

		return imports;
	}

	// @pre pkg != null
	public boolean addType(UserType type) {
    	types.put(type.getSimpleName(), type);
        getOwnerPackage().addType(type);
        return true;
    }

	public Set<UserType> getTypes() {
    	return new HashSet<UserType>(types.values());
    }

	public TypeDecl lookupType(boolean widen, String typeName, ReferenceType looker) {
		assert widen; // Can only get to SourceFile by widening.

    	TypeDecl result = types.get(typeName);

        if (result == null)
        	result = getImportType(typeName);
			// Try single type imports in this scope, but NOT "on demand" imports.
			// They happen after the package scope is searched, so are below.

		if (result != null)
        	return result.accessible(looker)? result : null;

		// Widen.

		result = getOwnerScope().lookupType(true, typeName, looker);

		if (result == null)
			result = getDemandType(typeName, looker);

		return result;
	}

	protected UserType getImportType(String typeName) {

		for (UserTypeLookup imported: importTypes) {
			UserType importedUserType = imported.getUserType();
            if (importedUserType != null && importedUserType.getSimpleName().equals(typeName))
                return importedUserType;
		}

		return null;
    }

	// Special lookup for on-demand imports.  This must be separate from normal type lookups,
    // because it happens after the package is searched.
	public TypeDecl getDemandType(String typeName, ReferenceType looker) {

		for (PackageOrTypeLookup imported: importsOnDemand) {
			Scope importedScope = imported.getScope();

			if (importedScope != null) {
				TypeDecl importedType = importedScope.lookupType(false, typeName, looker);
				if (importedType != null)
					return importedType;
			}
		}

		return null;
    }

	public MethodDecl lookupMethod(boolean widen, String methodName, List params, ReferenceType looker) {
    	return null; // Source files & their owners don't contain methods directly.
    }

	public VariableDecl lookupVariable(boolean widen, String variableName, ReferenceType looker) {
    	return null; // Source files & their owners don't contain variables directly.
    }

    public void accept(ModelVisitor visitor) {
        visitor.visitSourceFile(this);
    }
}
