package parser.parsetree;

import parser.grammar.*;
import parser.parsetree.visitor.*;

import java.util.*;

/**
* Generic syntax tree branches can be used for any grammar.
* They contain the production type, which indicates which production was recognised to find this branch.
* This library does not contain any non-generic branch classes because they are grammar-specific.
* E.g. we might generate an AdditionExpression node for a certain grammar.
*/
public class ParseTreeBranch extends SyntaxTreeArrayBranch {

    /**
     * The production that was recognised to build this branch.
     * May be null, in which case this node is ambiguous.
     *
     * @supplierCardinality 1
     * @clientCardinality 0..* 
     */
	protected Production productionType;

    /** Construct a GenericSyntaxTreeBranch.
     * 
	 *  @param theProductionType The type of the production represented by this leaf node.
     */
    public ParseTreeBranch(Production theProductionType, List theKids) {
        this(theProductionType, (SyntaxTree[]) theKids.toArray(new SyntaxTree[0]));
    }

    protected ParseTreeBranch(Production theProductionType, SyntaxTree[] theKids) {
		super(theKids);

        assert kids != null;
        //assert theProductionType != null;
        //assert kids.length == theProductionType.getSize();

        productionType = theProductionType;
    }

    /**
     * Override equals() because GLR parsers can repeat some reductions, and we don't
     * want identical "ambiguous" subtrees.
     */
    /*
    public boolean equals(Object o) {

        if (working)
            return true;

        if (!(o instanceof ParseTreeBranch))
            return false;

        ParseTreeBranch that = (ParseTreeBranch) o;
        if (productionType != that.getProductionType())
            return false;

        if (kids.length != that.getKids().length)
            return false;

        for (int i = 0; i < kids.length; i++) {
            if (kids[i] == null) {
                if (that.getKids()[i] != null)
                    return false;
            }
            else {
                working = true;
                boolean equalKid = kids[i].equals(that.getKids()[i]);
                working = false;
                if (!equalKid)
                    return false;
            }
        }

        return true;
    }
    */

    /** Keep hashcode consistent with equals. */
    /*
    public int hashCode() {
        if (working)
            return 0;

        int result = 0;
        if (productionType != null)
            result = productionType.hashCode() * 31;

        working = true;
        for (int i = 0; i < kids.length; i++)
            result = 31 * result + (kids[i] == null ? 0 : kids[i].hashCode());

        working = false;

        return result;
    }
    */

    /** Getter for the nonterminal on the lhs of the matched production. */
    public Nonterminal getNonterminalType() {
		if (productionType != null)
            return productionType.getLhs();

        return  ((AbstractSyntaxTreeBranch) kids[0]).getNonterminalType();  // All kids must have same type.
    }

    /** Getter for the production recognised. */
	public Production getProductionType() {
	    return productionType;
	}

    /** Get the string value of the tree.  I.e. the contents of the leaves. */
    public String getValue() {
        if (productionType != null)
            return super.getValue();

        return  kids[0].getValue();
    }

    /**
     * Tell this node it is now (more) ambiguous, & supply the alternative.
     * @param alternative a subtree that competes with this one.
     */
    public void ambiguity(SyntaxTreeBranch alternate) {
        assert getSymbolType() == alternate.getSymbolType();

        if (!isAmbiguous()) { // First ambiguity
            ParseTreeBranch original = new ParseTreeBranch(productionType, kids);
            kids = new SyntaxTree[1];
            kids[0] = original;

            productionType = null; // Flag ambiguity
        }

        SyntaxTree[] newKids = new SyntaxTree[kids.length + 1];

        for (int i = 0; i < kids.length; i++)
            newKids[i] = kids[i];

        newKids[kids.length] = alternate;

        kids = newKids;
    }

    public boolean isAmbiguous() {
        return productionType == null;
    }

    public boolean isList() {
        return false;
    }

	public String toString() {
        return productionType + " (" + getLine() + ", " + getColumn() + ")";
    }
}
