package parser.parsetree.treeFactory;

import parser.grammar.*;
import parser.parsetree.*;

import java.util.*;

/**
* Syntax tree branch factory that builds Abstract Syntax Trees that more closely
* match EBNF productions.  It supports optional rhs symbols by inserting nulls
* for missing symbols.  It supports lists by converting recursive trees into
* a list.
* <p>
* Normal productions are represented in the usual way using GenericSyntaxTreeBranches.
*/
public class ASTFactory implements SyntaxTreeFactory {

    /** Maps BNF productions to EBNF ones containing nulls. */
	protected Map optionals; // Production ==> Production

    /** Describes the position to insert nulls in a BNF production. */
	protected Map optionalTemplates;

    /** Nonterminals that define a left-recursive list. */
	protected Set listNonterminals; // NonterminalSymbols

    /* Construct a factory, with no special productions by default. */
	public ASTFactory() {
		optionals = new HashMap();
		optionalTemplates = new HashMap();
		listNonterminals = new HashSet();
	}

    /**
     * Specify a BNF production to be replaced with an EBNF one, and where to insert nulls.
     *
     * @param original the BNF production
     * @param replacement the EBNF production (same structure as production, with some nulls)
     * @param template array of booleans matching the length of the replacement production --
     *                 true indicates a value is provided in the BNF production,
     *                 false indicates a null should be inserted.
     */
	public void optional(Production original, Production replacement, boolean[] template) {
		optionals.put(original, replacement);
		optionalTemplates.put(original, template);
	}

    /**
     * Specify the BNF productions that define a left-recursive list.
     * E.g.  "X ::= X Y" and "X ::= Y".
     *
     * @param lhs the nonterminal on the left of a left-recursive list.
     */
	public void list(Nonterminal lhs) {
		listNonterminals.add(lhs);
	}

    /**
     * Create a branch node, adjusting structure as defined by optional() and list() settings.
     *
     * @param type the Type of production recognised.
     * @param kids the child nodes of this branch.
     * @return a new branch node.
     */
    public SyntaxTreeBranch makeBranch(Production prod, List kids) {
        assert kids.size() == prod.getSize();

        if (optionals.containsKey(prod))
            return makeOptionalBranch(prod, kids);

        if (listNonterminals.contains(prod.getLhs()))
            return makeList(prod, kids);

        // Normal node.
        return branchFactory(prod, kids);
    }

    /**
     * Helper method to construct a branch, for the case where a BNF production
     * should have nulls inserted to become an EBNF branch.
     */
    protected SyntaxTreeBranch makeOptionalBranch(Production type, List kids) {
        Production replacement = (Production) optionals.get(type);

        boolean[] template = (boolean[]) optionalTemplates.get(type);
        assert template != null;
        assert template.length >= kids.size();

        List paddedKids = new ArrayList(template.length);

        int kidPos = 0;
        for (int i = 0; i < template.length; i++) {
            if (template[i])
                paddedKids.add(kids.get(kidPos++));
            else
                paddedKids.add(null);
        }

        assert kidPos == kids.size();

        // Replace artificial BNF production with EBNF one containing some nulls.
        return branchFactory(replacement, paddedKids);
    }

    /**
     * Helper function to construct a branch.
     */
    protected SyntaxTreeBranch makeList(Production type, List kids) {
        // Replace artificial BNF production with AST list.
        assert kids.size() < 3;

        if (kids.size() == 1)
            return new ASTList(type, (SyntaxTree) kids.get(0));
        else {
            // We know the left kid must already be an ASTList. (Left recursion.)
            ASTList left = (ASTList) kids.get(0);
            return new ASTList(type, left, (SyntaxTree) kids.get(1));
        }
    }

    protected SyntaxTreeBranch branchFactory(Production type, List kids) {
        return new ParseTreeBranch(type, kids);
    }

    /**
     * Make an existing node ambiguous.  Checks first whether the alternative is identical.
     *
     * @param existing a parse tree node that may already be in use in a tree.
     * @param alternative an ambiguous alternative that has the same lhs.
     */
	public void makeAmbiguity(SyntaxTreeBranch existing, SyntaxTreeBranch alternative) {
        assert false;
    }
}
